from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Empty(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class Manifest(_message.Message):
    __slots__ = ("name", "version", "description", "author", "homepage", "license", "capabilities", "analyzer_rules_json")
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    AUTHOR_FIELD_NUMBER: _ClassVar[int]
    HOMEPAGE_FIELD_NUMBER: _ClassVar[int]
    LICENSE_FIELD_NUMBER: _ClassVar[int]
    CAPABILITIES_FIELD_NUMBER: _ClassVar[int]
    ANALYZER_RULES_JSON_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    description: str
    author: str
    homepage: str
    license: str
    capabilities: _containers.RepeatedScalarFieldContainer[str]
    analyzer_rules_json: bytes
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ..., description: _Optional[str] = ..., author: _Optional[str] = ..., homepage: _Optional[str] = ..., license: _Optional[str] = ..., capabilities: _Optional[_Iterable[str]] = ..., analyzer_rules_json: _Optional[bytes] = ...) -> None: ...

class ActionRequest(_message.Message):
    __slots__ = ("action_name", "trace_id", "data_json", "config_json", "context_json", "app_id", "app_name")
    ACTION_NAME_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    DATA_JSON_FIELD_NUMBER: _ClassVar[int]
    CONFIG_JSON_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_JSON_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    APP_NAME_FIELD_NUMBER: _ClassVar[int]
    action_name: str
    trace_id: str
    data_json: bytes
    config_json: bytes
    context_json: bytes
    app_id: str
    app_name: str
    def __init__(self, action_name: _Optional[str] = ..., trace_id: _Optional[str] = ..., data_json: _Optional[bytes] = ..., config_json: _Optional[bytes] = ..., context_json: _Optional[bytes] = ..., app_id: _Optional[str] = ..., app_name: _Optional[str] = ...) -> None: ...

class ActionResponse(_message.Message):
    __slots__ = ("success", "result_json", "error", "error_code")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    RESULT_JSON_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    result_json: bytes
    error: str
    error_code: str
    def __init__(self, success: bool = ..., result_json: _Optional[bytes] = ..., error: _Optional[str] = ..., error_code: _Optional[str] = ...) -> None: ...

class ActionMeta(_message.Message):
    __slots__ = ("name", "category", "description", "tags", "inputs_json", "outputs_json", "examples_json")
    NAME_FIELD_NUMBER: _ClassVar[int]
    CATEGORY_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    INPUTS_JSON_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_JSON_FIELD_NUMBER: _ClassVar[int]
    EXAMPLES_JSON_FIELD_NUMBER: _ClassVar[int]
    name: str
    category: str
    description: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    inputs_json: bytes
    outputs_json: bytes
    examples_json: bytes
    def __init__(self, name: _Optional[str] = ..., category: _Optional[str] = ..., description: _Optional[str] = ..., tags: _Optional[_Iterable[str]] = ..., inputs_json: _Optional[bytes] = ..., outputs_json: _Optional[bytes] = ..., examples_json: _Optional[bytes] = ...) -> None: ...

class ActionList(_message.Message):
    __slots__ = ("actions",)
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    actions: _containers.RepeatedCompositeFieldContainer[ActionMeta]
    def __init__(self, actions: _Optional[_Iterable[_Union[ActionMeta, _Mapping]]] = ...) -> None: ...

class ProviderConfig(_message.Message):
    __slots__ = ("name", "config_json")
    NAME_FIELD_NUMBER: _ClassVar[int]
    CONFIG_JSON_FIELD_NUMBER: _ClassVar[int]
    name: str
    config_json: bytes
    def __init__(self, name: _Optional[str] = ..., config_json: _Optional[bytes] = ...) -> None: ...

class ProviderName(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class ProviderResponse(_message.Message):
    __slots__ = ("success", "error")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error: str
    def __init__(self, success: bool = ..., error: _Optional[str] = ...) -> None: ...

class HealthResponse(_message.Message):
    __slots__ = ("healthy", "message", "details")
    class DetailsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    HEALTHY_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    healthy: bool
    message: str
    details: _containers.ScalarMap[str, str]
    def __init__(self, healthy: bool = ..., message: _Optional[str] = ..., details: _Optional[_Mapping[str, str]] = ...) -> None: ...

class PluginStatus(_message.Message):
    __slots__ = ("manifest", "uptime_seconds", "resources", "stats")
    MANIFEST_FIELD_NUMBER: _ClassVar[int]
    UPTIME_SECONDS_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_FIELD_NUMBER: _ClassVar[int]
    STATS_FIELD_NUMBER: _ClassVar[int]
    manifest: Manifest
    uptime_seconds: int
    resources: _containers.RepeatedCompositeFieldContainer[PluginResource]
    stats: PluginStats
    def __init__(self, manifest: _Optional[_Union[Manifest, _Mapping]] = ..., uptime_seconds: _Optional[int] = ..., resources: _Optional[_Iterable[_Union[PluginResource, _Mapping]]] = ..., stats: _Optional[_Union[PluginStats, _Mapping]] = ...) -> None: ...

class PluginResource(_message.Message):
    __slots__ = ("type", "name", "status", "info")
    class InfoEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    TYPE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    INFO_FIELD_NUMBER: _ClassVar[int]
    type: str
    name: str
    status: str
    info: _containers.ScalarMap[str, str]
    def __init__(self, type: _Optional[str] = ..., name: _Optional[str] = ..., status: _Optional[str] = ..., info: _Optional[_Mapping[str, str]] = ...) -> None: ...

class PluginStats(_message.Message):
    __slots__ = ("total_executions", "successful", "failed", "avg_latency_ms", "action_counts")
    class ActionCountsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: int
        def __init__(self, key: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...
    TOTAL_EXECUTIONS_FIELD_NUMBER: _ClassVar[int]
    SUCCESSFUL_FIELD_NUMBER: _ClassVar[int]
    FAILED_FIELD_NUMBER: _ClassVar[int]
    AVG_LATENCY_MS_FIELD_NUMBER: _ClassVar[int]
    ACTION_COUNTS_FIELD_NUMBER: _ClassVar[int]
    total_executions: int
    successful: int
    failed: int
    avg_latency_ms: float
    action_counts: _containers.ScalarMap[str, int]
    def __init__(self, total_executions: _Optional[int] = ..., successful: _Optional[int] = ..., failed: _Optional[int] = ..., avg_latency_ms: _Optional[float] = ..., action_counts: _Optional[_Mapping[str, int]] = ...) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import plugin_pb2 as plugin__pb2

GRPC_GENERATED_VERSION = '1.76.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in plugin_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class PluginServiceStub(object):
    """PluginService - Every plugin implements this service
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetManifest = channel.unary_unary(
                '/app4.plugin.PluginService/GetManifest',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.Manifest.FromString,
                _registered_method=True)
        self.ExecuteAction = channel.unary_unary(
                '/app4.plugin.PluginService/ExecuteAction',
                request_serializer=plugin__pb2.ActionRequest.SerializeToString,
                response_deserializer=plugin__pb2.ActionResponse.FromString,
                _registered_method=True)
        self.ListActions = channel.unary_unary(
                '/app4.plugin.PluginService/ListActions',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.ActionList.FromString,
                _registered_method=True)
        self.InitProvider = channel.unary_unary(
                '/app4.plugin.PluginService/InitProvider',
                request_serializer=plugin__pb2.ProviderConfig.SerializeToString,
                response_deserializer=plugin__pb2.ProviderResponse.FromString,
                _registered_method=True)
        self.CloseProvider = channel.unary_unary(
                '/app4.plugin.PluginService/CloseProvider',
                request_serializer=plugin__pb2.ProviderName.SerializeToString,
                response_deserializer=plugin__pb2.ProviderResponse.FromString,
                _registered_method=True)
        self.Health = channel.unary_unary(
                '/app4.plugin.PluginService/Health',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.HealthResponse.FromString,
                _registered_method=True)
        self.Shutdown = channel.unary_unary(
                '/app4.plugin.PluginService/Shutdown',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.GetStatus = channel.unary_unary(
                '/app4.plugin.PluginService/GetStatus',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.PluginStatus.FromString,
                _registered_method=True)


class PluginServiceServicer(object):
    """PluginService - Every plugin implements this service
    """

    def GetManifest(self, request, context):
        """Plugin identity and capabilities
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecuteAction(self, request, context):
        """Action execution
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListActions(self, request, context):
        """List all actions provided by this plugin
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InitProvider(self, request, context):
        """Provider lifecycle
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CloseProvider(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Health(self, request, context):
        """Health check
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Shutdown(self, request, context):
        """Graceful shutdown
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStatus(self, request, context):
        """Detailed plugin status (resources, stats, uptime)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PluginServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetManifest': grpc.unary_unary_rpc_method_handler(
                    servicer.GetManifest,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.Manifest.SerializeToString,
            ),
            'ExecuteAction': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecuteAction,
                    request_deserializer=plugin__pb2.ActionRequest.FromString,
                    response_serializer=plugin__pb2.ActionResponse.SerializeToString,
            ),
            'ListActions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListActions,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.ActionList.SerializeToString,
            ),
            'InitProvider': grpc.unary_unary_rpc_method_handler(
                    servicer.InitProvider,
                    request_deserializer=plugin__pb2.ProviderConfig.FromString,
                    response_serializer=plugin__pb2.ProviderResponse.SerializeToString,
            ),
            'CloseProvider': grpc.unary_unary_rpc_method_handler(
                    servicer.CloseProvider,
                    request_deserializer=plugin__pb2.ProviderName.FromString,
                    response_serializer=plugin__pb2.ProviderResponse.SerializeToString,
            ),
            'Health': grpc.unary_unary_rpc_method_handler(
                    servicer.Health,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.HealthResponse.SerializeToString,
            ),
            'Shutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.Shutdown,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'GetStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStatus,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.PluginStatus.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'app4.plugin.PluginService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('app4.plugin.PluginService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class PluginService(object):
    """PluginService - Every plugin implements this service
    """

    @staticmethod
    def GetManifest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/GetManifest',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.Manifest.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExecuteAction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/ExecuteAction',
            plugin__pb2.ActionRequest.SerializeToString,
            plugin__pb2.ActionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListActions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/ListActions',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.ActionList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def InitProvider(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/InitProvider',
            plugin__pb2.ProviderConfig.SerializeToString,
            plugin__pb2.ProviderResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CloseProvider(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/CloseProvider',
            plugin__pb2.ProviderName.SerializeToString,
            plugin__pb2.ProviderResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Health(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/Health',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.HealthResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Shutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/Shutdown',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.plugin.PluginService/GetStatus',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.PluginStatus.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: plugin.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'plugin.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cplugin.proto\x12\x0b\x61pp4.plugin\"\x07\n\x05\x45mpty\"\xa4\x01\n\x08Manifest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x04 \x01(\t\x12\x10\n\x08homepage\x18\x05 \x01(\t\x12\x0f\n\x07license\x18\x06 \x01(\t\x12\x14\n\x0c\x63\x61pabilities\x18\x07 \x03(\t\x12\x1b\n\x13\x61nalyzer_rules_json\x18\x08 \x01(\x0c\"\x96\x01\n\rActionRequest\x12\x13\n\x0b\x61\x63tion_name\x18\x01 \x01(\t\x12\x10\n\x08trace_id\x18\x02 \x01(\t\x12\x11\n\tdata_json\x18\x03 \x01(\x0c\x12\x13\n\x0b\x63onfig_json\x18\x04 \x01(\x0c\x12\x14\n\x0c\x63ontext_json\x18\x05 \x01(\x0c\x12\x0e\n\x06\x61pp_id\x18\x06 \x01(\t\x12\x10\n\x08\x61pp_name\x18\x07 \x01(\t\"Y\n\x0e\x41\x63tionResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x13\n\x0bresult_json\x18\x02 \x01(\x0c\x12\r\n\x05\x65rror\x18\x03 \x01(\t\x12\x12\n\nerror_code\x18\x04 \x01(\t\"\x91\x01\n\nActionMeta\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x63\x61tegory\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0c\n\x04tags\x18\x04 \x03(\t\x12\x13\n\x0binputs_json\x18\x05 \x01(\x0c\x12\x14\n\x0coutputs_json\x18\x06 \x01(\x0c\x12\x15\n\rexamples_json\x18\x07 \x01(\x0c\"6\n\nActionList\x12(\n\x07\x61\x63tions\x18\x01 \x03(\x0b\x32\x17.app4.plugin.ActionMeta\"3\n\x0eProviderConfig\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x63onfig_json\x18\x02 \x01(\x0c\"\x1c\n\x0cProviderName\x12\x0c\n\x04name\x18\x01 \x01(\t\"2\n\x10ProviderResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"\x9d\x01\n\x0eHealthResponse\x12\x0f\n\x07healthy\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x39\n\x07\x64\x65tails\x18\x03 \x03(\x0b\x32(.app4.plugin.HealthResponse.DetailsEntry\x1a.\n\x0c\x44\x65tailsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa8\x01\n\x0cPluginStatus\x12\'\n\x08manifest\x18\x01 \x01(\x0b\x32\x15.app4.plugin.Manifest\x12\x16\n\x0euptime_seconds\x18\x02 \x01(\x03\x12.\n\tresources\x18\x03 \x03(\x0b\x32\x1b.app4.plugin.PluginResource\x12\'\n\x05stats\x18\x04 \x01(\x0b\x32\x18.app4.plugin.PluginStats\"\x9e\x01\n\x0ePluginResource\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x33\n\x04info\x18\x04 \x03(\x0b\x32%.app4.plugin.PluginResource.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xdb\x01\n\x0bPluginStats\x12\x18\n\x10total_executions\x18\x01 \x01(\x03\x12\x12\n\nsuccessful\x18\x02 \x01(\x03\x12\x0e\n\x06\x66\x61iled\x18\x03 \x01(\x03\x12\x16\n\x0e\x61vg_latency_ms\x18\x04 \x01(\x01\x12\x41\n\raction_counts\x18\x05 \x03(\x0b\x32*.app4.plugin.PluginStats.ActionCountsEntry\x1a\x33\n\x11\x41\x63tionCountsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\x32\x91\x04\n\rPluginService\x12\x38\n\x0bGetManifest\x12\x12.app4.plugin.Empty\x1a\x15.app4.plugin.Manifest\x12H\n\rExecuteAction\x12\x1a.app4.plugin.ActionRequest\x1a\x1b.app4.plugin.ActionResponse\x12:\n\x0bListActions\x12\x12.app4.plugin.Empty\x1a\x17.app4.plugin.ActionList\x12J\n\x0cInitProvider\x12\x1b.app4.plugin.ProviderConfig\x1a\x1d.app4.plugin.ProviderResponse\x12I\n\rCloseProvider\x12\x19.app4.plugin.ProviderName\x1a\x1d.app4.plugin.ProviderResponse\x12\x39\n\x06Health\x12\x12.app4.plugin.Empty\x1a\x1b.app4.plugin.HealthResponse\x12\x32\n\x08Shutdown\x12\x12.app4.plugin.Empty\x1a\x12.app4.plugin.Empty\x12:\n\tGetStatus\x12\x12.app4.plugin.Empty\x1a\x19.app4.plugin.PluginStatusB,Z*gitlab.com/app4ai/app4-plugin-sdk/go/protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'plugin_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z*gitlab.com/app4ai/app4-plugin-sdk/go/proto'
  _globals['_HEALTHRESPONSE_DETAILSENTRY']._loaded_options = None
  _globals['_HEALTHRESPONSE_DETAILSENTRY']._serialized_options = b'8\001'
  _globals['_PLUGINRESOURCE_INFOENTRY']._loaded_options = None
  _globals['_PLUGINRESOURCE_INFOENTRY']._serialized_options = b'8\001'
  _globals['_PLUGINSTATS_ACTIONCOUNTSENTRY']._loaded_options = None
  _globals['_PLUGINSTATS_ACTIONCOUNTSENTRY']._serialized_options = b'8\001'
  _globals['_EMPTY']._serialized_start=29
  _globals['_EMPTY']._serialized_end=36
  _globals['_MANIFEST']._serialized_start=39
  _globals['_MANIFEST']._serialized_end=203
  _globals['_ACTIONREQUEST']._serialized_start=206
  _globals['_ACTIONREQUEST']._serialized_end=356
  _globals['_ACTIONRESPONSE']._serialized_start=358
  _globals['_ACTIONRESPONSE']._serialized_end=447
  _globals['_ACTIONMETA']._serialized_start=450
  _globals['_ACTIONMETA']._serialized_end=595
  _globals['_ACTIONLIST']._serialized_start=597
  _globals['_ACTIONLIST']._serialized_end=651
  _globals['_PROVIDERCONFIG']._serialized_start=653
  _globals['_PROVIDERCONFIG']._serialized_end=704
  _globals['_PROVIDERNAME']._serialized_start=706
  _globals['_PROVIDERNAME']._serialized_end=734
  _globals['_PROVIDERRESPONSE']._serialized_start=736
  _globals['_PROVIDERRESPONSE']._serialized_end=786
  _globals['_HEALTHRESPONSE']._serialized_start=789
  _globals['_HEALTHRESPONSE']._serialized_end=946
  _globals['_HEALTHRESPONSE_DETAILSENTRY']._serialized_start=900
  _globals['_HEALTHRESPONSE_DETAILSENTRY']._serialized_end=946
  _globals['_PLUGINSTATUS']._serialized_start=949
  _globals['_PLUGINSTATUS']._serialized_end=1117
  _globals['_PLUGINRESOURCE']._serialized_start=1120
  _globals['_PLUGINRESOURCE']._serialized_end=1278
  _globals['_PLUGINRESOURCE_INFOENTRY']._serialized_start=1235
  _globals['_PLUGINRESOURCE_INFOENTRY']._serialized_end=1278
  _globals['_PLUGINSTATS']._serialized_start=1281
  _globals['_PLUGINSTATS']._serialized_end=1500
  _globals['_PLUGINSTATS_ACTIONCOUNTSENTRY']._serialized_start=1449
  _globals['_PLUGINSTATS_ACTIONCOUNTSENTRY']._serialized_end=1500
  _globals['_PLUGINSERVICE']._serialized_start=1503
  _globals['_PLUGINSERVICE']._serialized_end=2032
# @@protoc_insertion_point(module_scope)



import plugin_pb2 as _plugin_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RegisterRequest(_message.Message):
    __slots__ = ("plugin_name", "version", "address", "instance_id", "labels")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    version: str
    address: str
    instance_id: str
    labels: _containers.ScalarMap[str, str]
    def __init__(self, plugin_name: _Optional[str] = ..., version: _Optional[str] = ..., address: _Optional[str] = ..., instance_id: _Optional[str] = ..., labels: _Optional[_Mapping[str, str]] = ...) -> None: ...

class RegisterResponse(_message.Message):
    __slots__ = ("success", "error", "assigned_id")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    ASSIGNED_ID_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error: str
    assigned_id: str
    def __init__(self, success: bool = ..., error: _Optional[str] = ..., assigned_id: _Optional[str] = ...) -> None: ...

class UnregisterRequest(_message.Message):
    __slots__ = ("plugin_name", "version", "instance_id")
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    version: str
    instance_id: str
    def __init__(self, plugin_name: _Optional[str] = ..., version: _Optional[str] = ..., instance_id: _Optional[str] = ...) -> None: ...

class HeartbeatRequest(_message.Message):
    __slots__ = ("plugin_name", "version", "instance_id", "healthy", "metrics")
    class MetricsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    HEALTHY_FIELD_NUMBER: _ClassVar[int]
    METRICS_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    version: str
    instance_id: str
    healthy: bool
    metrics: _containers.ScalarMap[str, str]
    def __init__(self, plugin_name: _Optional[str] = ..., version: _Optional[str] = ..., instance_id: _Optional[str] = ..., healthy: bool = ..., metrics: _Optional[_Mapping[str, str]] = ...) -> None: ...

class HeartbeatResponse(_message.Message):
    __slots__ = ("acknowledged", "should_shutdown", "deprecated", "deprecation_message", "recommended_version")
    ACKNOWLEDGED_FIELD_NUMBER: _ClassVar[int]
    SHOULD_SHUTDOWN_FIELD_NUMBER: _ClassVar[int]
    DEPRECATED_FIELD_NUMBER: _ClassVar[int]
    DEPRECATION_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    RECOMMENDED_VERSION_FIELD_NUMBER: _ClassVar[int]
    acknowledged: bool
    should_shutdown: bool
    deprecated: bool
    deprecation_message: str
    recommended_version: str
    def __init__(self, acknowledged: bool = ..., should_shutdown: bool = ..., deprecated: bool = ..., deprecation_message: _Optional[str] = ..., recommended_version: _Optional[str] = ...) -> None: ...

class ResolveRequest(_message.Message):
    __slots__ = ("plugin_name", "version_constraint", "prefer_labels")
    class PreferLabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_CONSTRAINT_FIELD_NUMBER: _ClassVar[int]
    PREFER_LABELS_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    version_constraint: str
    prefer_labels: _containers.ScalarMap[str, str]
    def __init__(self, plugin_name: _Optional[str] = ..., version_constraint: _Optional[str] = ..., prefer_labels: _Optional[_Mapping[str, str]] = ...) -> None: ...

class ResolveResponse(_message.Message):
    __slots__ = ("found", "resolved_version", "address", "instance_id", "error")
    FOUND_FIELD_NUMBER: _ClassVar[int]
    RESOLVED_VERSION_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    found: bool
    resolved_version: str
    address: str
    instance_id: str
    error: str
    def __init__(self, found: bool = ..., resolved_version: _Optional[str] = ..., address: _Optional[str] = ..., instance_id: _Optional[str] = ..., error: _Optional[str] = ...) -> None: ...

class ListVersionsRequest(_message.Message):
    __slots__ = ("plugin_name", "include_unhealthy")
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_UNHEALTHY_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    include_unhealthy: bool
    def __init__(self, plugin_name: _Optional[str] = ..., include_unhealthy: bool = ...) -> None: ...

class ListVersionsResponse(_message.Message):
    __slots__ = ("versions",)
    VERSIONS_FIELD_NUMBER: _ClassVar[int]
    versions: _containers.RepeatedCompositeFieldContainer[PluginVersion]
    def __init__(self, versions: _Optional[_Iterable[_Union[PluginVersion, _Mapping]]] = ...) -> None: ...

class PluginVersion(_message.Message):
    __slots__ = ("version", "instances", "deprecated", "deprecation_message", "recommended_version", "deprecated_at")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    INSTANCES_FIELD_NUMBER: _ClassVar[int]
    DEPRECATED_FIELD_NUMBER: _ClassVar[int]
    DEPRECATION_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    RECOMMENDED_VERSION_FIELD_NUMBER: _ClassVar[int]
    DEPRECATED_AT_FIELD_NUMBER: _ClassVar[int]
    version: str
    instances: _containers.RepeatedCompositeFieldContainer[PluginInstance]
    deprecated: bool
    deprecation_message: str
    recommended_version: str
    deprecated_at: int
    def __init__(self, version: _Optional[str] = ..., instances: _Optional[_Iterable[_Union[PluginInstance, _Mapping]]] = ..., deprecated: bool = ..., deprecation_message: _Optional[str] = ..., recommended_version: _Optional[str] = ..., deprecated_at: _Optional[int] = ...) -> None: ...

class PluginInstance(_message.Message):
    __slots__ = ("instance_id", "address", "healthy", "last_heartbeat", "labels", "metrics")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class MetricsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    HEALTHY_FIELD_NUMBER: _ClassVar[int]
    LAST_HEARTBEAT_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    METRICS_FIELD_NUMBER: _ClassVar[int]
    instance_id: str
    address: str
    healthy: bool
    last_heartbeat: int
    labels: _containers.ScalarMap[str, str]
    metrics: _containers.ScalarMap[str, str]
    def __init__(self, instance_id: _Optional[str] = ..., address: _Optional[str] = ..., healthy: bool = ..., last_heartbeat: _Optional[int] = ..., labels: _Optional[_Mapping[str, str]] = ..., metrics: _Optional[_Mapping[str, str]] = ...) -> None: ...

class SubscribeRequest(_message.Message):
    __slots__ = ("plugin_names",)
    PLUGIN_NAMES_FIELD_NUMBER: _ClassVar[int]
    plugin_names: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, plugin_names: _Optional[_Iterable[str]] = ...) -> None: ...

class RegistryEvent(_message.Message):
    __slots__ = ("type", "plugin_name", "version", "instance_id", "address", "timestamp")
    class EventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        PLUGIN_REGISTERED: _ClassVar[RegistryEvent.EventType]
        PLUGIN_UNREGISTERED: _ClassVar[RegistryEvent.EventType]
        PLUGIN_HEALTHY: _ClassVar[RegistryEvent.EventType]
        PLUGIN_UNHEALTHY: _ClassVar[RegistryEvent.EventType]
        VERSION_DEPRECATED: _ClassVar[RegistryEvent.EventType]
    PLUGIN_REGISTERED: RegistryEvent.EventType
    PLUGIN_UNREGISTERED: RegistryEvent.EventType
    PLUGIN_HEALTHY: RegistryEvent.EventType
    PLUGIN_UNHEALTHY: RegistryEvent.EventType
    VERSION_DEPRECATED: RegistryEvent.EventType
    TYPE_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    type: RegistryEvent.EventType
    plugin_name: str
    version: str
    instance_id: str
    address: str
    timestamp: int
    def __init__(self, type: _Optional[_Union[RegistryEvent.EventType, str]] = ..., plugin_name: _Optional[str] = ..., version: _Optional[str] = ..., instance_id: _Optional[str] = ..., address: _Optional[str] = ..., timestamp: _Optional[int] = ...) -> None: ...

class RegistryHealthResponse(_message.Message):
    __slots__ = ("healthy", "total_plugins", "healthy_plugins", "total_instances", "healthy_instances")
    HEALTHY_FIELD_NUMBER: _ClassVar[int]
    TOTAL_PLUGINS_FIELD_NUMBER: _ClassVar[int]
    HEALTHY_PLUGINS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    HEALTHY_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    healthy: bool
    total_plugins: int
    healthy_plugins: int
    total_instances: int
    healthy_instances: int
    def __init__(self, healthy: bool = ..., total_plugins: _Optional[int] = ..., healthy_plugins: _Optional[int] = ..., total_instances: _Optional[int] = ..., healthy_instances: _Optional[int] = ...) -> None: ...

class PluginList(_message.Message):
    __slots__ = ("plugins",)
    PLUGINS_FIELD_NUMBER: _ClassVar[int]
    plugins: _containers.RepeatedCompositeFieldContainer[PluginInfo]
    def __init__(self, plugins: _Optional[_Iterable[_Union[PluginInfo, _Mapping]]] = ...) -> None: ...

class PluginInfo(_message.Message):
    __slots__ = ("name", "available_versions", "total_instances", "healthy_instances")
    NAME_FIELD_NUMBER: _ClassVar[int]
    AVAILABLE_VERSIONS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    HEALTHY_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    name: str
    available_versions: _containers.RepeatedScalarFieldContainer[str]
    total_instances: int
    healthy_instances: int
    def __init__(self, name: _Optional[str] = ..., available_versions: _Optional[_Iterable[str]] = ..., total_instances: _Optional[int] = ..., healthy_instances: _Optional[int] = ...) -> None: ...

class DeprecateVersionRequest(_message.Message):
    __slots__ = ("plugin_name", "version", "message", "recommended_version")
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    RECOMMENDED_VERSION_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    version: str
    message: str
    recommended_version: str
    def __init__(self, plugin_name: _Optional[str] = ..., version: _Optional[str] = ..., message: _Optional[str] = ..., recommended_version: _Optional[str] = ...) -> None: ...

class DeprecateVersionResponse(_message.Message):
    __slots__ = ("success", "error", "affected_instances")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    AFFECTED_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error: str
    affected_instances: int
    def __init__(self, success: bool = ..., error: _Optional[str] = ..., affected_instances: _Optional[int] = ...) -> None: ...

class RemoveVersionRequest(_message.Message):
    __slots__ = ("plugin_name", "version", "force")
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    FORCE_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    version: str
    force: bool
    def __init__(self, plugin_name: _Optional[str] = ..., version: _Optional[str] = ..., force: bool = ...) -> None: ...

class RemoveVersionResponse(_message.Message):
    __slots__ = ("success", "error", "removed_instances")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    REMOVED_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error: str
    removed_instances: int
    def __init__(self, success: bool = ..., error: _Optional[str] = ..., removed_instances: _Optional[int] = ...) -> None: ...

class RegistryProviderConfig(_message.Message):
    __slots__ = ("app_id", "plugin_name", "provider_name", "config_json", "version", "created_by", "created_at", "updated_at")
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_NAME_FIELD_NUMBER: _ClassVar[int]
    CONFIG_JSON_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    plugin_name: str
    provider_name: str
    config_json: bytes
    version: int
    created_by: str
    created_at: int
    updated_at: int
    def __init__(self, app_id: _Optional[str] = ..., plugin_name: _Optional[str] = ..., provider_name: _Optional[str] = ..., config_json: _Optional[bytes] = ..., version: _Optional[int] = ..., created_by: _Optional[str] = ..., created_at: _Optional[int] = ..., updated_at: _Optional[int] = ...) -> None: ...

class StoreProviderConfigRequest(_message.Message):
    __slots__ = ("app_id", "plugin_name", "provider_name", "config_json", "created_by")
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_NAME_FIELD_NUMBER: _ClassVar[int]
    CONFIG_JSON_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    plugin_name: str
    provider_name: str
    config_json: bytes
    created_by: str
    def __init__(self, app_id: _Optional[str] = ..., plugin_name: _Optional[str] = ..., provider_name: _Optional[str] = ..., config_json: _Optional[bytes] = ..., created_by: _Optional[str] = ...) -> None: ...

class StoreProviderConfigResponse(_message.Message):
    __slots__ = ("success", "error", "version")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error: str
    version: int
    def __init__(self, success: bool = ..., error: _Optional[str] = ..., version: _Optional[int] = ...) -> None: ...

class GetProviderConfigRequest(_message.Message):
    __slots__ = ("app_id", "plugin_name", "provider_name")
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_NAME_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    plugin_name: str
    provider_name: str
    def __init__(self, app_id: _Optional[str] = ..., plugin_name: _Optional[str] = ..., provider_name: _Optional[str] = ...) -> None: ...

class GetProviderConfigResponse(_message.Message):
    __slots__ = ("found", "config", "error")
    FOUND_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    found: bool
    config: RegistryProviderConfig
    error: str
    def __init__(self, found: bool = ..., config: _Optional[_Union[RegistryProviderConfig, _Mapping]] = ..., error: _Optional[str] = ...) -> None: ...

class DeleteProviderConfigRequest(_message.Message):
    __slots__ = ("app_id", "plugin_name", "provider_name")
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_NAME_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    plugin_name: str
    provider_name: str
    def __init__(self, app_id: _Optional[str] = ..., plugin_name: _Optional[str] = ..., provider_name: _Optional[str] = ...) -> None: ...

class ListProviderConfigsRequest(_message.Message):
    __slots__ = ("app_id", "plugin_name")
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    plugin_name: str
    def __init__(self, app_id: _Optional[str] = ..., plugin_name: _Optional[str] = ...) -> None: ...

class ListProviderConfigsResponse(_message.Message):
    __slots__ = ("configs",)
    CONFIGS_FIELD_NUMBER: _ClassVar[int]
    configs: _containers.RepeatedCompositeFieldContainer[RegistryProviderConfig]
    def __init__(self, configs: _Optional[_Iterable[_Union[RegistryProviderConfig, _Mapping]]] = ...) -> None: ...

class SubscribeProviderConfigsRequest(_message.Message):
    __slots__ = ("app_id", "plugin_name")
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    plugin_name: str
    def __init__(self, app_id: _Optional[str] = ..., plugin_name: _Optional[str] = ...) -> None: ...

class ProviderConfigEvent(_message.Message):
    __slots__ = ("type", "app_id", "plugin_name", "provider_name", "version", "timestamp")
    class EventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        CONFIG_CREATED: _ClassVar[ProviderConfigEvent.EventType]
        CONFIG_UPDATED: _ClassVar[ProviderConfigEvent.EventType]
        CONFIG_DELETED: _ClassVar[ProviderConfigEvent.EventType]
    CONFIG_CREATED: ProviderConfigEvent.EventType
    CONFIG_UPDATED: ProviderConfigEvent.EventType
    CONFIG_DELETED: ProviderConfigEvent.EventType
    TYPE_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    type: ProviderConfigEvent.EventType
    app_id: str
    plugin_name: str
    provider_name: str
    version: int
    timestamp: int
    def __init__(self, type: _Optional[_Union[ProviderConfigEvent.EventType, str]] = ..., app_id: _Optional[str] = ..., plugin_name: _Optional[str] = ..., provider_name: _Optional[str] = ..., version: _Optional[int] = ..., timestamp: _Optional[int] = ...) -> None: ...

class ConnectionEvent(_message.Message):
    __slots__ = ("plugin_instance_id", "plugin_name", "plugin_version", "app_id", "app_name", "timestamp")
    PLUGIN_INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_VERSION_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    APP_NAME_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    plugin_instance_id: str
    plugin_name: str
    plugin_version: str
    app_id: str
    app_name: str
    timestamp: int
    def __init__(self, plugin_instance_id: _Optional[str] = ..., plugin_name: _Optional[str] = ..., plugin_version: _Optional[str] = ..., app_id: _Optional[str] = ..., app_name: _Optional[str] = ..., timestamp: _Optional[int] = ...) -> None: ...

class ListConnectionsRequest(_message.Message):
    __slots__ = ("plugin_name", "app_id", "include_history")
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_HISTORY_FIELD_NUMBER: _ClassVar[int]
    plugin_name: str
    app_id: str
    include_history: bool
    def __init__(self, plugin_name: _Optional[str] = ..., app_id: _Optional[str] = ..., include_history: bool = ...) -> None: ...

class ListConnectionsResponse(_message.Message):
    __slots__ = ("connections",)
    CONNECTIONS_FIELD_NUMBER: _ClassVar[int]
    connections: _containers.RepeatedCompositeFieldContainer[AppConnection]
    def __init__(self, connections: _Optional[_Iterable[_Union[AppConnection, _Mapping]]] = ...) -> None: ...

class AppConnection(_message.Message):
    __slots__ = ("app_id", "app_name", "plugin_name", "plugin_version", "plugin_instance_id", "plugin_address", "connected_at", "last_activity", "active")
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    APP_NAME_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_VERSION_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_ADDRESS_FIELD_NUMBER: _ClassVar[int]
    CONNECTED_AT_FIELD_NUMBER: _ClassVar[int]
    LAST_ACTIVITY_FIELD_NUMBER: _ClassVar[int]
    ACTIVE_FIELD_NUMBER: _ClassVar[int]
    app_id: str
    app_name: str
    plugin_name: str
    plugin_version: str
    plugin_instance_id: str
    plugin_address: str
    connected_at: int
    last_activity: int
    active: bool
    def __init__(self, app_id: _Optional[str] = ..., app_name: _Optional[str] = ..., plugin_name: _Optional[str] = ..., plugin_version: _Optional[str] = ..., plugin_instance_id: _Optional[str] = ..., plugin_address: _Optional[str] = ..., connected_at: _Optional[int] = ..., last_activity: _Optional[int] = ..., active: bool = ...) -> None: ...



"""
Registry client for App4 plugins.

This module provides the RegistryClient class for interacting with
the plugin registry service.
"""

import logging
import threading
import time
import uuid
from dataclasses import dataclass, field
from typing import Any, Callable, Optional

import grpc

# Import generated proto modules
try:
    from app4_sdk.proto import plugin_pb2
    from app4_sdk.proto import registry_pb2
    from app4_sdk.proto import registry_pb2_grpc
except ImportError:
    plugin_pb2 = None
    registry_pb2 = None
    registry_pb2_grpc = None


logger = logging.getLogger("app4.plugin.registry")


@dataclass
class RegistryConfig:
    """Configuration for registry client."""
    address: str
    heartbeat_seconds: int = 10
    instance_id: Optional[str] = None
    labels: dict[str, str] = field(default_factory=dict)


@dataclass
class ConfigEvent:
    """Event for provider config changes."""
    type: str  # "created", "updated", "deleted"
    app_id: str
    plugin_name: str
    provider_name: str
    version: int
    timestamp: int


class RegistryClient:
    """
    Client for interacting with the plugin registry.

    Handles:
    - Plugin registration and unregistration
    - Periodic heartbeats
    - App connection tracking
    - Provider config management
    """

    def __init__(
        self,
        config: RegistryConfig,
        plugin_name: str,
        version: str,
        address: str,
    ):
        """
        Create a new registry client.

        Args:
            config: Registry configuration
            plugin_name: Name of the plugin
            version: Version of the plugin
            address: gRPC address this plugin is listening on
        """
        self.config = config
        self.plugin_name = plugin_name
        self.version = version
        self.address = address
        self.instance_id = config.instance_id or str(uuid.uuid4())

        self._channel: Optional[grpc.Channel] = None
        self._stub: Optional[Any] = None
        self._heartbeat_thread: Optional[threading.Thread] = None
        self._running = False
        self._registered = False

    def start(self) -> None:
        """
        Start the registry client.

        Connects to the registry, registers the plugin, and starts
        the heartbeat thread.
        """
        if registry_pb2 is None:
            raise RuntimeError("gRPC proto modules not generated")

        # Connect to registry
        self._channel = grpc.insecure_channel(self.config.address)
        self._stub = registry_pb2_grpc.PluginRegistryServiceStub(self._channel)

        # Register
        try:
            request = registry_pb2.RegisterRequest(
                plugin_name=self.plugin_name,
                version=self.version,
                address=self.address,
                instance_id=self.instance_id,
                labels=self.config.labels,
            )
            response = self._stub.Register(request)

            if response.success:
                self._registered = True
                if response.assigned_id:
                    self.instance_id = response.assigned_id
                logger.info(f"Registered with registry: {self.instance_id}")
            else:
                raise RuntimeError(f"Registration failed: {response.error}")

        except grpc.RpcError as e:
            logger.error(f"Failed to register: {e}")
            raise

        # Start heartbeat thread
        self._running = True
        self._heartbeat_thread = threading.Thread(target=self._heartbeat_loop, daemon=True)
        self._heartbeat_thread.start()

    def stop(self) -> None:
        """
        Stop the registry client.

        Stops the heartbeat thread and unregisters from the registry.
        """
        self._running = False

        if self._heartbeat_thread:
            self._heartbeat_thread.join(timeout=2)

        if self._registered and self._stub:
            try:
                request = registry_pb2.UnregisterRequest(
                    plugin_name=self.plugin_name,
                    version=self.version,
                    instance_id=self.instance_id,
                )
                self._stub.Unregister(request)
                logger.info("Unregistered from registry")
            except grpc.RpcError as e:
                logger.warning(f"Failed to unregister: {e}")

        if self._channel:
            self._channel.close()

    def _heartbeat_loop(self) -> None:
        """Background thread for sending heartbeats."""
        while self._running:
            try:
                self._send_heartbeat()
            except Exception as e:
                logger.warning(f"Heartbeat failed: {e}")

            # Sleep in small increments to allow quick shutdown
            for _ in range(self.config.heartbeat_seconds * 10):
                if not self._running:
                    break
                time.sleep(0.1)

    def _send_heartbeat(self) -> None:
        """Send a heartbeat to the registry."""
        if not self._stub:
            return

        request = registry_pb2.HeartbeatRequest(
            plugin_name=self.plugin_name,
            version=self.version,
            instance_id=self.instance_id,
            healthy=True,
        )

        try:
            response = self._stub.Heartbeat(request)

            if response.should_shutdown:
                logger.warning("Registry requested shutdown")
                self._running = False

            if response.deprecated:
                logger.warning(
                    f"This version is deprecated: {response.deprecation_message}. "
                    f"Recommended version: {response.recommended_version}"
                )

        except grpc.RpcError as e:
            logger.warning(f"Heartbeat RPC failed: {e}")

    def notify_connect(self, app_id: str, app_name: str = "") -> None:
        """
        Notify the registry that an app has connected.

        Args:
            app_id: Application ID
            app_name: Application name (optional)
        """
        if not self._stub:
            return

        try:
            request = registry_pb2.ConnectionEvent(
                plugin_instance_id=self.instance_id,
                plugin_name=self.plugin_name,
                plugin_version=self.version,
                app_id=app_id,
                app_name=app_name,
                timestamp=int(time.time()),
            )
            self._stub.NotifyConnect(request)
            logger.debug(f"Notified connect: {app_id}")

        except grpc.RpcError as e:
            logger.warning(f"Failed to notify connect: {e}")

    def notify_disconnect(self, app_id: str) -> None:
        """
        Notify the registry that an app has disconnected.

        Args:
            app_id: Application ID
        """
        if not self._stub:
            return

        try:
            request = registry_pb2.ConnectionEvent(
                plugin_instance_id=self.instance_id,
                plugin_name=self.plugin_name,
                plugin_version=self.version,
                app_id=app_id,
                timestamp=int(time.time()),
            )
            self._stub.NotifyDisconnect(request)
            logger.debug(f"Notified disconnect: {app_id}")

        except grpc.RpcError as e:
            logger.warning(f"Failed to notify disconnect: {e}")


class ConfigClient:
    """
    Client for managing provider configurations via the registry.

    Provides caching and change notification for provider configs.
    """

    def __init__(self, registry_address: str):
        """
        Create a config client.

        Args:
            registry_address: Registry gRPC address
        """
        self.registry_address = registry_address
        self._channel: Optional[grpc.Channel] = None
        self._stub: Optional[Any] = None
        self._cache: dict[str, dict[str, Any]] = {}  # key -> {config, version, cached_at}
        self._event_handlers: list[Callable[[ConfigEvent], None]] = []
        self._subscription_thread: Optional[threading.Thread] = None
        self._running = False

    def connect(self) -> None:
        """Connect to the registry."""
        if registry_pb2 is None:
            raise RuntimeError("gRPC proto modules not generated")

        self._channel = grpc.insecure_channel(self.registry_address)
        self._stub = registry_pb2_grpc.PluginRegistryServiceStub(self._channel)

    def close(self) -> None:
        """Close the connection."""
        self._running = False
        if self._subscription_thread:
            self._subscription_thread.join(timeout=2)
        if self._channel:
            self._channel.close()

    def _cache_key(self, app_id: str, plugin_name: str, provider_name: str) -> str:
        """Generate cache key."""
        return f"{app_id}:{plugin_name}:{provider_name}"

    def get_config(
        self,
        app_id: str,
        plugin_name: str,
        provider_name: str,
    ) -> Optional[dict[str, Any]]:
        """
        Get provider configuration.

        Returns cached config if available, otherwise fetches from registry.

        Args:
            app_id: Application ID
            plugin_name: Plugin name
            provider_name: Provider instance name

        Returns:
            Config dictionary or None if not found
        """
        if not self._stub:
            self.connect()

        key = self._cache_key(app_id, plugin_name, provider_name)

        # Check cache
        if key in self._cache:
            return self._cache[key]["config"]

        # Fetch from registry
        try:
            request = registry_pb2.GetProviderConfigRequest(
                app_id=app_id,
                plugin_name=plugin_name,
                provider_name=provider_name,
            )
            response = self._stub.GetProviderConfig(request)

            if response.found:
                import json
                config = json.loads(response.config.config_json)
                self._cache[key] = {
                    "config": config,
                    "version": response.config.version,
                    "cached_at": time.time(),
                }
                return config

        except grpc.RpcError as e:
            logger.warning(f"Failed to get config: {e}")

        return None

    def store_config(
        self,
        app_id: str,
        plugin_name: str,
        provider_name: str,
        config: dict[str, Any],
        created_by: str = "",
    ) -> bool:
        """
        Store provider configuration.

        Args:
            app_id: Application ID
            plugin_name: Plugin name
            provider_name: Provider instance name
            config: Configuration dictionary
            created_by: Identifier of the creator

        Returns:
            True if successful
        """
        if not self._stub:
            self.connect()

        try:
            import json
            request = registry_pb2.StoreProviderConfigRequest(
                app_id=app_id,
                plugin_name=plugin_name,
                provider_name=provider_name,
                config_json=json.dumps(config).encode(),
                created_by=created_by,
            )
            response = self._stub.StoreProviderConfig(request)

            if response.success:
                # Update cache
                key = self._cache_key(app_id, plugin_name, provider_name)
                self._cache[key] = {
                    "config": config,
                    "version": response.version,
                    "cached_at": time.time(),
                }
                return True
            else:
                logger.warning(f"Failed to store config: {response.error}")
                return False

        except grpc.RpcError as e:
            logger.warning(f"Failed to store config: {e}")
            return False

    def subscribe(
        self,
        handler: Callable[[ConfigEvent], None],
        app_id: str = "",
        plugin_name: str = "",
    ) -> None:
        """
        Subscribe to config changes.

        Args:
            handler: Callback function for config events
            app_id: Filter by app ID (empty for all)
            plugin_name: Filter by plugin name (empty for all)
        """
        if not self._stub:
            self.connect()

        self._event_handlers.append(handler)

        if not self._subscription_thread:
            self._running = True
            self._subscription_thread = threading.Thread(
                target=self._subscription_loop,
                args=(app_id, plugin_name),
                daemon=True,
            )
            self._subscription_thread.start()

    def _subscription_loop(self, app_id: str, plugin_name: str) -> None:
        """Background thread for subscription."""
        while self._running:
            try:
                request = registry_pb2.SubscribeProviderConfigsRequest(
                    app_id=app_id,
                    plugin_name=plugin_name,
                )

                for event in self._stub.SubscribeProviderConfigs(request):
                    if not self._running:
                        break

                    config_event = ConfigEvent(
                        type=["created", "updated", "deleted"][event.type],
                        app_id=event.app_id,
                        plugin_name=event.plugin_name,
                        provider_name=event.provider_name,
                        version=event.version,
                        timestamp=event.timestamp,
                    )

                    # Invalidate cache
                    key = self._cache_key(
                        event.app_id, event.plugin_name, event.provider_name
                    )
                    if key in self._cache:
                        del self._cache[key]

                    # Notify handlers
                    for handler in self._event_handlers:
                        try:
                            handler(config_event)
                        except Exception as e:
                            logger.warning(f"Handler error: {e}")

            except grpc.RpcError as e:
                if self._running:
                    logger.warning(f"Subscription error: {e}, reconnecting...")
                    time.sleep(5)



"""
Export utilities for App4 plugins.

This module provides functions to export plugin metadata and
analyzer rules in various formats.
"""

import json
import sys
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from app4_sdk.plugin.plugin import Plugin


def _format_value(v: Any) -> str:
    """Format a value for YAML output."""
    if isinstance(v, str):
        # Check if it's a reference (starts with $ or ~)
        if v.startswith("$") or v.startswith("~"):
            return v
        return f'"{v}"'
    elif isinstance(v, dict):
        return json.dumps(v)
    elif isinstance(v, list):
        return json.dumps(v)
    else:
        return str(v)


def export_metadata(plugin: "Plugin") -> None:
    """
    Export action metadata as JSON to stdout.

    Args:
        plugin: Plugin instance
    """
    actions = plugin.list_actions()
    manifest = plugin.manifest()

    output = {
        "plugin": manifest.to_dict(),
        "actions": [a.to_dict() for a in actions],
    }

    print(json.dumps(output, indent=2))


def export_rules(plugin: "Plugin") -> None:
    """
    Export analyzer rules as JSON to stdout.

    Args:
        plugin: Plugin instance
    """
    rules = plugin.analyzer_rules()

    if rules:
        print(json.dumps(rules.to_dict(), indent=2))
    else:
        print("{}")


def export_all(plugin: "Plugin") -> None:
    """
    Export both metadata and rules as JSON to stdout.

    Args:
        plugin: Plugin instance
    """
    actions = plugin.list_actions()
    manifest = plugin.manifest()
    rules = plugin.analyzer_rules()

    output = {
        "plugin": manifest.to_dict(),
        "actions": [a.to_dict() for a in actions],
        "analyzerRules": rules.to_dict() if rules else None,
    }

    print(json.dumps(output, indent=2))


def export_markdown(plugin: "Plugin") -> None:
    """
    Export plugin documentation as AI-friendly Markdown to stdout.

    Args:
        plugin: Plugin instance
    """
    manifest = plugin.manifest()
    actions = plugin.list_actions()
    rules = plugin.analyzer_rules()

    # Build write actions set from rules
    write_actions_set = set()
    if rules and rules.write_actions:
        for action in rules.write_actions:
            write_actions_set.add(f"{manifest.name}:{action}")

    # Categorize actions
    write_actions = []
    read_actions = []
    for action in actions:
        if action.name in write_actions_set:
            write_actions.append(action)
        else:
            read_actions.append(action)

    lines = [
        f"# {manifest.name} Plugin",
        "",
        "<!-- AI-FRIENDLY: This document describes a plugin for the App4 meta-model system -->",
        "",
        "## Overview",
        "",
        manifest.description if manifest.description else "",
        "",
        "| Property | Value |",
        "|----------|-------|",
        f"| Name | `{manifest.name}` |",
        f"| Version | `{manifest.version}` |",
    ]

    # Add provider type if available from rules
    provider_type = manifest.name  # Default to plugin name
    if rules and hasattr(rules, 'provider_type') and rules.provider_type:
        provider_type = rules.provider_type
    lines.append(f"| Provider | `$provider:{provider_type}:<name>` |")
    lines.append("")

    # Actions summary by category
    lines.append("## Actions")
    lines.append("")

    if write_actions:
        lines.extend([
            "### Write Actions",
            "",
            "| Action | Description |",
            "|--------|-------------|",
        ])
        for action in write_actions:
            desc = action.description if action.description else ""
            lines.append(f"| `{action.name}` | {desc} |")
        lines.append("")

    if read_actions:
        lines.extend([
            "### Read Actions",
            "",
            "| Action | Description |",
            "|--------|-------------|",
        ])
        for action in read_actions:
            desc = action.description if action.description else ""
            lines.append(f"| `{action.name}` | {desc} |")
        lines.append("")

    # Detailed action reference
    lines.extend([
        "## Action Reference",
        "",
    ])

    for action in actions:
        lines.extend([
            f"### `{action.name}`",
            "",
            action.description if action.description else "",
            "",
        ])

        if action.tags:
            tags_str = ", ".join([f"`{t}`" for t in action.tags])
            lines.append(f"**Tags:** {tags_str}")
            lines.append("")

        if action.inputs:
            lines.extend([
                "**Configuration:**",
                "",
                "| Parameter | Type | Required | Default | Description |",
                "|-----------|------|----------|---------|-------------|",
            ])
            for name, param in action.inputs.items():
                required = "**Yes**" if param.required else "No"
                default = getattr(param, 'default', None)
                default_str = str(default) if default is not None else "-"
                desc = param.description.replace("|", "\\|") if param.description else ""
                lines.append(f"| `{name}` | {param.type} | {required} | {default_str} | {desc} |")
            lines.append("")

        # Examples section
        if hasattr(action, 'examples') and action.examples:
            lines.append("**Examples:**")
            lines.append("")
            for ex in action.examples:
                ex_title = getattr(ex, 'title', 'Example')
                ex_desc = getattr(ex, 'description', '')
                lines.append(f"#### {ex_title}")
                lines.append("")
                if ex_desc:
                    lines.append(ex_desc)
                    lines.append("")

                # JSON config example
                if hasattr(ex, 'config') and ex.config:
                    lines.append("**JSON Config:**")
                    lines.append("```yaml")
                    lines.append(f"- action: {action.name}")
                    lines.append("  config:")
                    for k, v in ex.config.items():
                        lines.append(f"    {k}: {_format_value(v)}")
                    lines.append("```")
                    lines.append("")

                # DSL examples
                if hasattr(ex, 'dsl_examples') and ex.dsl_examples:
                    lines.append("**DSL Format:**")
                    lines.append("```app4")
                    for dsl in ex.dsl_examples:
                        lines.append(dsl)
                    lines.append("```")
                    lines.append("")
        else:
            # Fallback: generate basic example from inputs
            lines.extend([
                "**Example:**",
                "",
                "```yaml",
                f"- action: {action.name}",
                "  config:",
                f"    provider: $provider:{provider_type}:default",
            ])
            # Add required inputs as example
            if action.inputs:
                for name, param in action.inputs.items():
                    if name == "provider" or name == "target":
                        continue
                    if param.required:
                        if param.type == "string":
                            lines.append(f'    {name}: "value"')
                        elif param.type == "object":
                            lines.append(f"    {name}: {{}}")
                        elif param.type == "array":
                            lines.append(f"    {name}: []")
                        else:
                            lines.append(f"    {name}: # {param.type}")
            lines.append("```")
            lines.append("")

        lines.extend([
            "---",
            "",
        ])

    # Footer
    lines.extend([
        "---",
        "",
        "*Generated by App4 Plugin SDK*",
    ])

    print("\n".join(lines))


def handle_export_flags(plugin: "Plugin") -> bool:
    """
    Check for and handle export flags.

    Call this in your main() before serve() to support:
    - --export-meta: Export action metadata as JSON
    - --export-rules: Export analyzer rules as JSON
    - --export-all: Export both
    - --export-md: Export as Markdown documentation

    Args:
        plugin: Plugin instance

    Returns:
        True if an export flag was handled (should exit)
        False if no export flag present (continue to serve)

    Example:
        if __name__ == "__main__":
            plugin = MyPlugin()
            if handle_export_flags(plugin):
                sys.exit(0)
            serve(plugin)
    """
    if len(sys.argv) < 2:
        return False

    flag = sys.argv[1]

    if flag == "--export-meta":
        export_metadata(plugin)
        return True
    elif flag == "--export-rules":
        export_rules(plugin)
        return True
    elif flag == "--export-all":
        export_all(plugin)
        return True
    elif flag == "--export-md":
        export_markdown(plugin)
        return True

    return False



"""Plugin module for App4 SDK."""

from app4_sdk.plugin.types import (
    Manifest,
    ActionMeta,
    ParamSchema,
    ParamOption,
    ActionExample,
    AnalyzerRules,
    PatternSuggestion,
)
from app4_sdk.plugin.plugin import Plugin
from app4_sdk.plugin.context import Context, BaseContext
from app4_sdk.plugin.server import serve, serve_with_registry, serve_with_auto_registry
from app4_sdk.plugin.registry import RegistryClient, RegistryConfig
from app4_sdk.plugin.helpers import (
    get_config_string,
    get_config_bool,
    get_config_int,
    get_config_float,
    get_config_value,
    get_config_list,
    get_config_dict,
    require_config,
)
from app4_sdk.plugin.export import handle_export_flags

__all__ = [
    "Manifest",
    "ActionMeta",
    "ParamSchema",
    "ParamOption",
    "ActionExample",
    "AnalyzerRules",
    "PatternSuggestion",
    "Plugin",
    "Context",
    "BaseContext",
    "serve",
    "serve_with_registry",
    "serve_with_auto_registry",
    "RegistryClient",
    "RegistryConfig",
    "get_config_string",
    "get_config_bool",
    "get_config_int",
    "get_config_float",
    "get_config_value",
    "get_config_list",
    "get_config_dict",
    "require_config",
    "handle_export_flags",
]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from app4_sdk.proto import plugin_pb2 as plugin__pb2
import registry_pb2 as registry__pb2

GRPC_GENERATED_VERSION = '1.76.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in registry_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class PluginRegistryServiceStub(object):
    """PluginRegistryService - Central service for plugin version management
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Register = channel.unary_unary(
                '/app4.registry.PluginRegistryService/Register',
                request_serializer=registry__pb2.RegisterRequest.SerializeToString,
                response_deserializer=registry__pb2.RegisterResponse.FromString,
                _registered_method=True)
        self.Unregister = channel.unary_unary(
                '/app4.registry.PluginRegistryService/Unregister',
                request_serializer=registry__pb2.UnregisterRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.Heartbeat = channel.unary_unary(
                '/app4.registry.PluginRegistryService/Heartbeat',
                request_serializer=registry__pb2.HeartbeatRequest.SerializeToString,
                response_deserializer=registry__pb2.HeartbeatResponse.FromString,
                _registered_method=True)
        self.ResolveVersion = channel.unary_unary(
                '/app4.registry.PluginRegistryService/ResolveVersion',
                request_serializer=registry__pb2.ResolveRequest.SerializeToString,
                response_deserializer=registry__pb2.ResolveResponse.FromString,
                _registered_method=True)
        self.ListVersions = channel.unary_unary(
                '/app4.registry.PluginRegistryService/ListVersions',
                request_serializer=registry__pb2.ListVersionsRequest.SerializeToString,
                response_deserializer=registry__pb2.ListVersionsResponse.FromString,
                _registered_method=True)
        self.Subscribe = channel.unary_stream(
                '/app4.registry.PluginRegistryService/Subscribe',
                request_serializer=registry__pb2.SubscribeRequest.SerializeToString,
                response_deserializer=registry__pb2.RegistryEvent.FromString,
                _registered_method=True)
        self.Health = channel.unary_unary(
                '/app4.registry.PluginRegistryService/Health',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=registry__pb2.RegistryHealthResponse.FromString,
                _registered_method=True)
        self.ListAllPlugins = channel.unary_unary(
                '/app4.registry.PluginRegistryService/ListAllPlugins',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=registry__pb2.PluginList.FromString,
                _registered_method=True)
        self.DeprecateVersion = channel.unary_unary(
                '/app4.registry.PluginRegistryService/DeprecateVersion',
                request_serializer=registry__pb2.DeprecateVersionRequest.SerializeToString,
                response_deserializer=registry__pb2.DeprecateVersionResponse.FromString,
                _registered_method=True)
        self.RemoveVersion = channel.unary_unary(
                '/app4.registry.PluginRegistryService/RemoveVersion',
                request_serializer=registry__pb2.RemoveVersionRequest.SerializeToString,
                response_deserializer=registry__pb2.RemoveVersionResponse.FromString,
                _registered_method=True)
        self.StoreProviderConfig = channel.unary_unary(
                '/app4.registry.PluginRegistryService/StoreProviderConfig',
                request_serializer=registry__pb2.StoreProviderConfigRequest.SerializeToString,
                response_deserializer=registry__pb2.StoreProviderConfigResponse.FromString,
                _registered_method=True)
        self.GetProviderConfig = channel.unary_unary(
                '/app4.registry.PluginRegistryService/GetProviderConfig',
                request_serializer=registry__pb2.GetProviderConfigRequest.SerializeToString,
                response_deserializer=registry__pb2.GetProviderConfigResponse.FromString,
                _registered_method=True)
        self.DeleteProviderConfig = channel.unary_unary(
                '/app4.registry.PluginRegistryService/DeleteProviderConfig',
                request_serializer=registry__pb2.DeleteProviderConfigRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.ListProviderConfigs = channel.unary_unary(
                '/app4.registry.PluginRegistryService/ListProviderConfigs',
                request_serializer=registry__pb2.ListProviderConfigsRequest.SerializeToString,
                response_deserializer=registry__pb2.ListProviderConfigsResponse.FromString,
                _registered_method=True)
        self.SubscribeProviderConfigs = channel.unary_stream(
                '/app4.registry.PluginRegistryService/SubscribeProviderConfigs',
                request_serializer=registry__pb2.SubscribeProviderConfigsRequest.SerializeToString,
                response_deserializer=registry__pb2.ProviderConfigEvent.FromString,
                _registered_method=True)
        self.NotifyConnect = channel.unary_unary(
                '/app4.registry.PluginRegistryService/NotifyConnect',
                request_serializer=registry__pb2.ConnectionEvent.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.NotifyDisconnect = channel.unary_unary(
                '/app4.registry.PluginRegistryService/NotifyDisconnect',
                request_serializer=registry__pb2.ConnectionEvent.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.ListConnections = channel.unary_unary(
                '/app4.registry.PluginRegistryService/ListConnections',
                request_serializer=registry__pb2.ListConnectionsRequest.SerializeToString,
                response_deserializer=registry__pb2.ListConnectionsResponse.FromString,
                _registered_method=True)


class PluginRegistryServiceServicer(object):
    """PluginRegistryService - Central service for plugin version management
    """

    def Register(self, request, context):
        """Plugin registration (called by plugin instances on startup)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Unregister(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Heartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveVersion(self, request, context):
        """Version resolution (called by go-run instances)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListVersions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Subscribe(self, request, context):
        """Real-time updates
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Health(self, request, context):
        """Admin/health operations
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListAllPlugins(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeprecateVersion(self, request, context):
        """Version management
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StoreProviderConfig(self, request, context):
        """Provider config storage (shared state for multi-instance deployments)
        go-run stores configs here on startup, plugins fetch on cache miss
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetProviderConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteProviderConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListProviderConfigs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubscribeProviderConfigs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NotifyConnect(self, request, context):
        """App connection tracking (called by plugins when apps connect/disconnect)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NotifyDisconnect(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListConnections(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PluginRegistryServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Register': grpc.unary_unary_rpc_method_handler(
                    servicer.Register,
                    request_deserializer=registry__pb2.RegisterRequest.FromString,
                    response_serializer=registry__pb2.RegisterResponse.SerializeToString,
            ),
            'Unregister': grpc.unary_unary_rpc_method_handler(
                    servicer.Unregister,
                    request_deserializer=registry__pb2.UnregisterRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'Heartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.Heartbeat,
                    request_deserializer=registry__pb2.HeartbeatRequest.FromString,
                    response_serializer=registry__pb2.HeartbeatResponse.SerializeToString,
            ),
            'ResolveVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveVersion,
                    request_deserializer=registry__pb2.ResolveRequest.FromString,
                    response_serializer=registry__pb2.ResolveResponse.SerializeToString,
            ),
            'ListVersions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListVersions,
                    request_deserializer=registry__pb2.ListVersionsRequest.FromString,
                    response_serializer=registry__pb2.ListVersionsResponse.SerializeToString,
            ),
            'Subscribe': grpc.unary_stream_rpc_method_handler(
                    servicer.Subscribe,
                    request_deserializer=registry__pb2.SubscribeRequest.FromString,
                    response_serializer=registry__pb2.RegistryEvent.SerializeToString,
            ),
            'Health': grpc.unary_unary_rpc_method_handler(
                    servicer.Health,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=registry__pb2.RegistryHealthResponse.SerializeToString,
            ),
            'ListAllPlugins': grpc.unary_unary_rpc_method_handler(
                    servicer.ListAllPlugins,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=registry__pb2.PluginList.SerializeToString,
            ),
            'DeprecateVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.DeprecateVersion,
                    request_deserializer=registry__pb2.DeprecateVersionRequest.FromString,
                    response_serializer=registry__pb2.DeprecateVersionResponse.SerializeToString,
            ),
            'RemoveVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveVersion,
                    request_deserializer=registry__pb2.RemoveVersionRequest.FromString,
                    response_serializer=registry__pb2.RemoveVersionResponse.SerializeToString,
            ),
            'StoreProviderConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.StoreProviderConfig,
                    request_deserializer=registry__pb2.StoreProviderConfigRequest.FromString,
                    response_serializer=registry__pb2.StoreProviderConfigResponse.SerializeToString,
            ),
            'GetProviderConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetProviderConfig,
                    request_deserializer=registry__pb2.GetProviderConfigRequest.FromString,
                    response_serializer=registry__pb2.GetProviderConfigResponse.SerializeToString,
            ),
            'DeleteProviderConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteProviderConfig,
                    request_deserializer=registry__pb2.DeleteProviderConfigRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'ListProviderConfigs': grpc.unary_unary_rpc_method_handler(
                    servicer.ListProviderConfigs,
                    request_deserializer=registry__pb2.ListProviderConfigsRequest.FromString,
                    response_serializer=registry__pb2.ListProviderConfigsResponse.SerializeToString,
            ),
            'SubscribeProviderConfigs': grpc.unary_stream_rpc_method_handler(
                    servicer.SubscribeProviderConfigs,
                    request_deserializer=registry__pb2.SubscribeProviderConfigsRequest.FromString,
                    response_serializer=registry__pb2.ProviderConfigEvent.SerializeToString,
            ),
            'NotifyConnect': grpc.unary_unary_rpc_method_handler(
                    servicer.NotifyConnect,
                    request_deserializer=registry__pb2.ConnectionEvent.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'NotifyDisconnect': grpc.unary_unary_rpc_method_handler(
                    servicer.NotifyDisconnect,
                    request_deserializer=registry__pb2.ConnectionEvent.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'ListConnections': grpc.unary_unary_rpc_method_handler(
                    servicer.ListConnections,
                    request_deserializer=registry__pb2.ListConnectionsRequest.FromString,
                    response_serializer=registry__pb2.ListConnectionsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'app4.registry.PluginRegistryService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('app4.registry.PluginRegistryService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class PluginRegistryService(object):
    """PluginRegistryService - Central service for plugin version management
    """

    @staticmethod
    def Register(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/Register',
            registry__pb2.RegisterRequest.SerializeToString,
            registry__pb2.RegisterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Unregister(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/Unregister',
            registry__pb2.UnregisterRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Heartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/Heartbeat',
            registry__pb2.HeartbeatRequest.SerializeToString,
            registry__pb2.HeartbeatResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResolveVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/ResolveVersion',
            registry__pb2.ResolveRequest.SerializeToString,
            registry__pb2.ResolveResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListVersions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/ListVersions',
            registry__pb2.ListVersionsRequest.SerializeToString,
            registry__pb2.ListVersionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Subscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/app4.registry.PluginRegistryService/Subscribe',
            registry__pb2.SubscribeRequest.SerializeToString,
            registry__pb2.RegistryEvent.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Health(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/Health',
            plugin__pb2.Empty.SerializeToString,
            registry__pb2.RegistryHealthResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListAllPlugins(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/ListAllPlugins',
            plugin__pb2.Empty.SerializeToString,
            registry__pb2.PluginList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeprecateVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/DeprecateVersion',
            registry__pb2.DeprecateVersionRequest.SerializeToString,
            registry__pb2.DeprecateVersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RemoveVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/RemoveVersion',
            registry__pb2.RemoveVersionRequest.SerializeToString,
            registry__pb2.RemoveVersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StoreProviderConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/StoreProviderConfig',
            registry__pb2.StoreProviderConfigRequest.SerializeToString,
            registry__pb2.StoreProviderConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetProviderConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/GetProviderConfig',
            registry__pb2.GetProviderConfigRequest.SerializeToString,
            registry__pb2.GetProviderConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteProviderConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/DeleteProviderConfig',
            registry__pb2.DeleteProviderConfigRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListProviderConfigs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/ListProviderConfigs',
            registry__pb2.ListProviderConfigsRequest.SerializeToString,
            registry__pb2.ListProviderConfigsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SubscribeProviderConfigs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/app4.registry.PluginRegistryService/SubscribeProviderConfigs',
            registry__pb2.SubscribeProviderConfigsRequest.SerializeToString,
            registry__pb2.ProviderConfigEvent.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def NotifyConnect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/NotifyConnect',
            registry__pb2.ConnectionEvent.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def NotifyDisconnect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/NotifyDisconnect',
            registry__pb2.ConnectionEvent.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListConnections(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/app4.registry.PluginRegistryService/ListConnections',
            registry__pb2.ListConnectionsRequest.SerializeToString,
            registry__pb2.ListConnectionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""
App4 Plugin SDK for Python

A Python SDK for building App4 plugins that can integrate with the App4 platform
via gRPC protocol.

Example:
    from app4_sdk import Plugin, Manifest, ActionMeta, serve

    class MyPlugin(Plugin):
        def manifest(self) -> Manifest:
            return Manifest(name="my-plugin", version="1.0.0")

        def list_actions(self) -> list[ActionMeta]:
            return [...]

        def execute_action(self, ctx, action_name, data, config) -> dict:
            ...

    if __name__ == "__main__":
        serve(MyPlugin())
"""

from app4_sdk.plugin.types import (
    Manifest,
    ActionMeta,
    ParamSchema,
    ParamOption,
    ActionExample,
    AnalyzerRules,
    PatternSuggestion,
)
from app4_sdk.plugin.plugin import Plugin
from app4_sdk.plugin.context import Context, BaseContext
from app4_sdk.plugin.server import serve, serve_with_registry, serve_with_auto_registry
from app4_sdk.plugin.registry import RegistryClient, RegistryConfig
from app4_sdk.plugin.helpers import (
    get_config_string,
    get_config_bool,
    get_config_int,
    get_config_float,
    get_config_value,
    get_config_list,
    get_config_dict,
    require_config,
    get_nested_value,
    set_nested_value,
    is_namespaced_path,
    parse_action_name,
    copy_dict,
    merge_dicts,
)

__version__ = "1.0.0"
__all__ = [
    # Types
    "Manifest",
    "ActionMeta",
    "ParamSchema",
    "ParamOption",
    "ActionExample",
    "AnalyzerRules",
    "PatternSuggestion",
    # Plugin
    "Plugin",
    "Context",
    "BaseContext",
    # Server
    "serve",
    "serve_with_registry",
    "serve_with_auto_registry",
    # Registry
    "RegistryClient",
    "RegistryConfig",
    # Helpers
    "get_config_string",
    "get_config_bool",
    "get_config_int",
    "get_config_float",
    "get_config_value",
    "get_config_list",
    "get_config_dict",
    "require_config",
    "get_nested_value",
    "set_nested_value",
    "is_namespaced_path",
    "parse_action_name",
    "copy_dict",
    "merge_dicts",
]



"""
Execution context for plugin actions.

The Context provides access to:
- Logging methods (debug, info, warn, error)
- Input arguments ($.args.*)
- Context variables ($.ctx.*)
- Return values ($.return.*)
- Trace information
"""

from abc import ABC, abstractmethod
from typing import Any, Optional
import logging


class Context(ABC):
    """
    Abstract context interface for action execution.

    Provides logging, variable access, and trace information.
    """

    @abstractmethod
    def debug(self, msg: str, *args: Any) -> None:
        """Log a debug message."""
        pass

    @abstractmethod
    def info(self, msg: str, *args: Any) -> None:
        """Log an info message."""
        pass

    @abstractmethod
    def warn(self, msg: str, *args: Any) -> None:
        """Log a warning message."""
        pass

    @abstractmethod
    def error(self, msg: str, *args: Any) -> None:
        """Log an error message."""
        pass

    @abstractmethod
    def arg(self, name: str) -> Any:
        """Get an input argument ($.args.*)."""
        pass

    @abstractmethod
    def ctx(self, name: str) -> Any:
        """Get a context variable ($.ctx.*)."""
        pass

    @abstractmethod
    def set_ctx(self, name: str, value: Any) -> None:
        """Set a context variable ($.ctx.*)."""
        pass

    @abstractmethod
    def set_return(self, name: str, value: Any) -> None:
        """Set a return value ($.return.*)."""
        pass

    @property
    @abstractmethod
    def trace_id(self) -> str:
        """Get the trace ID for this execution."""
        pass

    @property
    @abstractmethod
    def step_name(self) -> str:
        """Get the current step name."""
        pass


class BaseContext(Context):
    """
    Base implementation of the execution context.

    Provides a simple in-memory implementation that can be used
    for testing or as a base for custom contexts.
    """

    def __init__(
        self,
        trace_id: str = "",
        step_name: str = "",
        args: Optional[dict[str, Any]] = None,
        ctx: Optional[dict[str, Any]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        self._trace_id = trace_id
        self._step_name = step_name
        self._args = args or {}
        self._ctx = ctx or {}
        self._return: dict[str, Any] = {}
        self._logger = logger or logging.getLogger("app4.plugin")

    def debug(self, msg: str, *args: Any) -> None:
        """Log a debug message."""
        self._logger.debug(f"[{self._trace_id}] {msg}", *args)

    def info(self, msg: str, *args: Any) -> None:
        """Log an info message."""
        self._logger.info(f"[{self._trace_id}] {msg}", *args)

    def warn(self, msg: str, *args: Any) -> None:
        """Log a warning message."""
        self._logger.warning(f"[{self._trace_id}] {msg}", *args)

    def error(self, msg: str, *args: Any) -> None:
        """Log an error message."""
        self._logger.error(f"[{self._trace_id}] {msg}", *args)

    def arg(self, name: str) -> Any:
        """Get an input argument ($.args.*)."""
        return self._args.get(name)

    def ctx(self, name: str) -> Any:
        """Get a context variable ($.ctx.*)."""
        return self._ctx.get(name)

    def set_ctx(self, name: str, value: Any) -> None:
        """Set a context variable ($.ctx.*)."""
        self._ctx[name] = value

    def set_return(self, name: str, value: Any) -> None:
        """Set a return value ($.return.*)."""
        self._return[name] = value

    @property
    def trace_id(self) -> str:
        """Get the trace ID for this execution."""
        return self._trace_id

    @property
    def step_name(self) -> str:
        """Get the current step name."""
        return self._step_name

    @property
    def return_values(self) -> dict[str, Any]:
        """Get all return values."""
        return self._return

    @property
    def context_values(self) -> dict[str, Any]:
        """Get all context values."""
        return self._ctx

    def to_dict(self) -> dict[str, Any]:
        """Convert context to dictionary for serialization."""
        return {
            "traceId": self._trace_id,
            "stepName": self._step_name,
            "args": self._args,
            "ctx": self._ctx,
            "return": self._return,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "BaseContext":
        """Create context from dictionary."""
        return cls(
            trace_id=data.get("traceId", ""),
            step_name=data.get("stepName", ""),
            args=data.get("args", {}),
            ctx=data.get("ctx", {}),
        )



"""
Plugin interface for App4 SDK.

This module defines the base Plugin class that all plugins must implement.
"""

from abc import ABC, abstractmethod
from typing import Any, Optional

from app4_sdk.plugin.types import (
    Manifest,
    ActionMeta,
    AnalyzerRules,
    PluginStatus,
    PluginResource,
    PluginStats,
)
from app4_sdk.plugin.context import Context


class Plugin(ABC):
    """
    Abstract base class for App4 plugins.

    All plugins must implement this interface to be compatible
    with the App4 platform.

    Example:
        class MyPlugin(Plugin):
            def manifest(self) -> Manifest:
                return Manifest(
                    name="my-plugin",
                    version="1.0.0",
                    description="My awesome plugin",
                    capabilities=["actions", "provider"],
                )

            def list_actions(self) -> list[ActionMeta]:
                return [
                    ActionMeta(
                        name="my-plugin:greet",
                        description="Say hello",
                        inputs={"name": ParamSchema(name="name", type="string")},
                        outputs={"message": ParamSchema(name="message", type="string")},
                    ),
                ]

            def execute_action(
                self, ctx: Context, action_name: str, data: dict, config: dict
            ) -> dict:
                if action_name == "my-plugin:greet":
                    name = data.get("name", "World")
                    return {"message": f"Hello, {name}!"}
                raise ValueError(f"Unknown action: {action_name}")

            def init_provider(self, name: str, config: dict) -> None:
                # Initialize provider connection
                pass

            def close_provider(self, name: str) -> None:
                # Close provider connection
                pass
    """

    @abstractmethod
    def manifest(self) -> Manifest:
        """
        Return the plugin manifest.

        The manifest describes the plugin's identity and capabilities.
        """
        pass

    @abstractmethod
    def list_actions(self) -> list[ActionMeta]:
        """
        Return a list of all actions provided by this plugin.

        Each action should have complete metadata including inputs,
        outputs, and examples.
        """
        pass

    @abstractmethod
    def execute_action(
        self,
        ctx: Context,
        action_name: str,
        data: dict[str, Any],
        config: dict[str, Any],
    ) -> dict[str, Any]:
        """
        Execute an action.

        Args:
            ctx: Execution context with logging and variable access
            action_name: Name of the action to execute (e.g., "mongo:find")
            data: Input data for the action
            config: Configuration for the action

        Returns:
            Result dictionary

        Raises:
            ValueError: If action is unknown
            Exception: If action execution fails
        """
        pass

    def init_provider(self, name: str, config: dict[str, Any]) -> None:
        """
        Initialize a provider instance.

        Override this method to set up connections or resources
        for a named provider instance.

        Args:
            name: Provider instance name (e.g., "default", "analytics")
            config: Provider configuration
        """
        pass

    def close_provider(self, name: str) -> None:
        """
        Close a provider instance.

        Override this method to clean up connections or resources
        for a named provider instance.

        Args:
            name: Provider instance name
        """
        pass

    def health_check(self) -> tuple[bool, str, dict[str, str]]:
        """
        Check plugin health.

        Override this method to provide custom health checking.

        Returns:
            Tuple of (healthy, message, details)
        """
        return True, "OK", {}

    def get_status(self) -> PluginStatus:
        """
        Get detailed plugin status.

        Override this method to provide custom status information
        including resources and statistics.

        Returns:
            PluginStatus with manifest, uptime, resources, and stats
        """
        return PluginStatus(
            manifest=self.manifest(),
            uptime_seconds=0,
            resources=[],
            stats=PluginStats(),
        )

    def analyzer_rules(self) -> Optional[AnalyzerRules]:
        """
        Return analyzer rules for this plugin.

        Override this method to provide validation rules for the
        meta-model analyzer.

        Returns:
            AnalyzerRules or None if no special rules
        """
        return None

    def shutdown(self) -> None:
        """
        Gracefully shutdown the plugin.

        Override this method to clean up all resources before shutdown.
        """
        pass



"""
gRPC server for App4 plugins.

This module provides functions to serve a plugin over gRPC:
- serve(): Simple server without registry
- serve_with_registry(): Server with registry integration
- serve_with_auto_registry(): Auto-configure registry from environment
"""

import json
import logging
import os
import signal
import sys
import time
from concurrent import futures
from typing import Optional

import grpc

from app4_sdk.plugin.plugin import Plugin
from app4_sdk.plugin.context import BaseContext
from app4_sdk.plugin.registry import RegistryClient, RegistryConfig

# Import generated proto modules
try:
    from app4_sdk.proto import plugin_pb2
    from app4_sdk.proto import plugin_pb2_grpc
except ImportError:
    # Proto modules not generated yet - provide stubs
    plugin_pb2 = None
    plugin_pb2_grpc = None


logger = logging.getLogger("app4.plugin.server")


class PluginServicer:
    """gRPC servicer implementation for Plugin interface."""

    def __init__(self, plugin: Plugin):
        self.plugin = plugin
        self.start_time = time.time()
        self._shutdown_requested = False

    def GetManifest(self, request, context):
        """Return plugin manifest."""
        manifest = self.plugin.manifest()
        rules = self.plugin.analyzer_rules()

        response = plugin_pb2.Manifest(
            name=manifest.name,
            version=manifest.version,
            description=manifest.description,
            author=manifest.author,
            homepage=manifest.homepage,
            license=manifest.license,
            capabilities=manifest.capabilities,
        )

        if rules:
            response.analyzer_rules_json = json.dumps(rules.to_dict()).encode()

        return response

    def ListActions(self, request, context):
        """Return list of all actions."""
        actions = self.plugin.list_actions()

        action_list = plugin_pb2.ActionList()
        for action in actions:
            action_meta = plugin_pb2.ActionMeta(
                name=action.name,
                category=action.category,
                description=action.description,
                tags=action.tags,
                inputs_json=json.dumps({k: v.to_dict() for k, v in action.inputs.items()}).encode(),
                outputs_json=json.dumps({k: v.to_dict() for k, v in action.outputs.items()}).encode(),
                examples_json=json.dumps([e.to_dict() for e in action.examples]).encode(),
            )
            action_list.actions.append(action_meta)

        return action_list

    def ExecuteAction(self, request, context):
        """Execute an action."""
        try:
            # Parse input data
            data = json.loads(request.data_json) if request.data_json else {}
            config = json.loads(request.config_json) if request.config_json else {}
            ctx_data = json.loads(request.context_json) if request.context_json else {}

            # Create context
            ctx = BaseContext(
                trace_id=request.trace_id,
                args=ctx_data.get("args", {}),
                ctx=ctx_data.get("ctx", {}),
            )

            # Execute action
            result = self.plugin.execute_action(ctx, request.action_name, data, config)

            return plugin_pb2.ActionResponse(
                success=True,
                result_json=json.dumps(result).encode(),
            )

        except Exception as e:
            logger.error(f"Action execution failed: {e}")
            error_code = type(e).__name__
            return plugin_pb2.ActionResponse(
                success=False,
                error=str(e),
                error_code=error_code,
            )

    def InitProvider(self, request, context):
        """Initialize a provider instance."""
        try:
            config = json.loads(request.config_json) if request.config_json else {}
            self.plugin.init_provider(request.name, config)
            return plugin_pb2.ProviderResponse(success=True)
        except Exception as e:
            logger.error(f"Provider init failed: {e}")
            return plugin_pb2.ProviderResponse(success=False, error=str(e))

    def CloseProvider(self, request, context):
        """Close a provider instance."""
        try:
            self.plugin.close_provider(request.name)
            return plugin_pb2.ProviderResponse(success=True)
        except Exception as e:
            logger.error(f"Provider close failed: {e}")
            return plugin_pb2.ProviderResponse(success=False, error=str(e))

    def Health(self, request, context):
        """Health check."""
        healthy, message, details = self.plugin.health_check()
        return plugin_pb2.HealthResponse(
            healthy=healthy,
            message=message,
            details=details,
        )

    def Shutdown(self, request, context):
        """Graceful shutdown."""
        self._shutdown_requested = True
        self.plugin.shutdown()
        return plugin_pb2.Empty()

    def GetStatus(self, request, context):
        """Get detailed plugin status."""
        status = self.plugin.get_status()

        # Build manifest proto
        manifest = plugin_pb2.Manifest(
            name=status.manifest.name,
            version=status.manifest.version,
            description=status.manifest.description,
            author=status.manifest.author,
            homepage=status.manifest.homepage,
            license=status.manifest.license,
            capabilities=status.manifest.capabilities,
        )

        # Build resources
        resources = []
        for r in status.resources:
            resources.append(plugin_pb2.PluginResource(
                type=r.type,
                name=r.name,
                status=r.status,
                info=r.info,
            ))

        # Build stats
        stats = None
        if status.stats:
            stats = plugin_pb2.PluginStats(
                total_executions=status.stats.total_executions,
                successful=status.stats.successful,
                failed=status.stats.failed,
                avg_latency_ms=status.stats.avg_latency_ms,
                action_counts=status.stats.action_counts,
            )

        # Calculate actual uptime
        uptime = int(time.time() - self.start_time)

        return plugin_pb2.PluginStatus(
            manifest=manifest,
            uptime_seconds=uptime,
            resources=resources,
            stats=stats,
        )


def serve(
    plugin: Plugin,
    address: str = ":50051",
    max_workers: int = 10,
) -> None:
    """
    Start a gRPC server for the plugin.

    This is the simplest way to serve a plugin without registry integration.

    Args:
        plugin: Plugin instance to serve
        address: gRPC address to listen on (default: ":50051")
        max_workers: Maximum thread pool workers (default: 10)

    Example:
        if __name__ == "__main__":
            plugin = MyPlugin()
            serve(plugin, ":50051")
    """
    if plugin_pb2 is None:
        raise RuntimeError(
            "gRPC proto modules not generated. Run: "
            "python -m grpc_tools.protoc -I./app4_sdk/proto "
            "--python_out=./app4_sdk/proto --grpc_python_out=./app4_sdk/proto "
            "./app4_sdk/proto/plugin.proto"
        )

    # Parse address
    if address.startswith(":"):
        address = f"0.0.0.0{address}"

    # Create server
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
    servicer = PluginServicer(plugin)
    plugin_pb2_grpc.add_PluginServiceServicer_to_server(servicer, server)

    server.add_insecure_port(address)

    # Setup signal handlers
    def signal_handler(signum, frame):
        logger.info("Shutdown signal received, stopping server...")
        server.stop(grace=5)
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    # Start server
    manifest = plugin.manifest()
    logger.info(f"Starting plugin {manifest.name} v{manifest.version} on {address}")
    server.start()

    try:
        server.wait_for_termination()
    except KeyboardInterrupt:
        logger.info("Interrupted, stopping server...")
        server.stop(grace=5)


def serve_with_registry(
    plugin: Plugin,
    address: str = ":50051",
    registry_config: Optional[RegistryConfig] = None,
    max_workers: int = 10,
) -> None:
    """
    Start a gRPC server with registry integration.

    The plugin will register with the registry on startup and send
    periodic heartbeats. On shutdown, it will unregister.

    Args:
        plugin: Plugin instance to serve
        address: gRPC address to listen on (default: ":50051")
        registry_config: Registry configuration
        max_workers: Maximum thread pool workers (default: 10)

    Example:
        if __name__ == "__main__":
            plugin = MyPlugin()
            config = RegistryConfig(address="localhost:50100")
            serve_with_registry(plugin, ":50051", config)
    """
    if plugin_pb2 is None:
        raise RuntimeError("gRPC proto modules not generated")

    # Parse address
    if address.startswith(":"):
        listen_address = f"0.0.0.0{address}"
    else:
        listen_address = address

    # Create server
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
    servicer = PluginServicer(plugin)
    plugin_pb2_grpc.add_PluginServiceServicer_to_server(servicer, server)

    server.add_insecure_port(listen_address)

    # Create registry client if configured
    registry_client = None
    if registry_config:
        manifest = plugin.manifest()
        registry_client = RegistryClient(
            config=registry_config,
            plugin_name=manifest.name,
            version=manifest.version,
            address=address,
        )

    # Setup signal handlers
    def signal_handler(signum, frame):
        logger.info("Shutdown signal received, stopping server...")
        if registry_client:
            registry_client.stop()
        server.stop(grace=5)
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    # Start server
    manifest = plugin.manifest()
    logger.info(f"Starting plugin {manifest.name} v{manifest.version} on {listen_address}")
    server.start()

    # Register with registry
    if registry_client:
        try:
            registry_client.start()
        except Exception as e:
            logger.error(f"Failed to register with registry: {e}")

    try:
        server.wait_for_termination()
    except KeyboardInterrupt:
        logger.info("Interrupted, stopping server...")
        if registry_client:
            registry_client.stop()
        server.stop(grace=5)


def serve_with_auto_registry(
    plugin: Plugin,
    address: str = ":50051",
    max_workers: int = 10,
) -> None:
    """
    Start a gRPC server with auto-configured registry.

    Registry configuration is read from environment variables:
    - PLUGIN_REGISTRY_ADDRESS: Registry gRPC address
    - PLUGIN_REGISTRY_HEARTBEAT: Heartbeat interval in seconds (default: 10)
    - PLUGIN_INSTANCE_ID: Custom instance ID (optional)

    If PLUGIN_REGISTRY_ADDRESS is not set, serves without registry.

    Args:
        plugin: Plugin instance to serve
        address: gRPC address to listen on (default: ":50051")
        max_workers: Maximum thread pool workers (default: 10)

    Example:
        if __name__ == "__main__":
            plugin = MyPlugin()
            # Set PLUGIN_REGISTRY_ADDRESS=localhost:50100
            serve_with_auto_registry(plugin, ":50051")
    """
    registry_address = os.environ.get("PLUGIN_REGISTRY_ADDRESS", "")

    if not registry_address:
        logger.info("No registry address configured, serving without registry")
        serve(plugin, address, max_workers)
        return

    heartbeat = int(os.environ.get("PLUGIN_REGISTRY_HEARTBEAT", "10"))
    instance_id = os.environ.get("PLUGIN_INSTANCE_ID", "")
    labels_str = os.environ.get("PLUGIN_LABELS", "")

    labels = {}
    if labels_str:
        for pair in labels_str.split(","):
            if "=" in pair:
                key, value = pair.split("=", 1)
                labels[key.strip()] = value.strip()

    config = RegistryConfig(
        address=registry_address,
        heartbeat_seconds=heartbeat,
        instance_id=instance_id if instance_id else None,
        labels=labels,
    )

    serve_with_registry(plugin, address, config, max_workers)


def check_export_flags() -> bool:
    """
    Check for export flags and handle them.

    Supports:
    - --export-meta: Export action metadata as JSON
    - --export-rules: Export analyzer rules as JSON
    - --export-all: Export both
    - --export-md: Export as Markdown documentation

    Returns:
        True if an export flag was handled, False otherwise
    """
    if len(sys.argv) < 2:
        return False

    flag = sys.argv[1]

    if flag not in ("--export-meta", "--export-rules", "--export-all", "--export-md"):
        return False

    # Import here to avoid circular imports
    from app4_sdk.plugin.export import export_metadata, export_rules, export_all, export_markdown

    if flag == "--export-meta":
        export_metadata()
    elif flag == "--export-rules":
        export_rules()
    elif flag == "--export-all":
        export_all()
    elif flag == "--export-md":
        export_markdown()

    return True



"""
Core types for App4 Plugin SDK.

This module defines all the data structures used throughout the SDK,
including plugin manifests, action metadata, and analyzer rules.
"""

from dataclasses import dataclass, field
from typing import Any, Optional
from enum import Enum


class UIComponent(str, Enum):
    """UI component types for action parameters."""
    TEXT_INPUT = "TextInput"
    TEXT_AREA = "TextArea"
    NUMBER_INPUT = "NumberInput"
    SELECT = "Select"
    MULTI_SELECT = "MultiSelect"
    CHECKBOX = "Checkbox"
    SWITCH = "Switch"
    DATE_PICKER = "DatePicker"
    TIME_PICKER = "TimePicker"
    COLOR_PICKER = "ColorPicker"
    FILE_UPLOAD = "FileUpload"
    CODE_EDITOR = "CodeEditor"
    JSON_EDITOR = "JsonEditor"
    SCHEMA_REF = "SchemaRef"
    PIPELINE_REF = "PipelineRef"
    PROVIDER_REF = "ProviderRef"


@dataclass
class ParamOption:
    """Option for select-type parameters."""
    label: str
    value: Any
    description: str = ""


@dataclass
class ParamSchema:
    """
    Parameter schema for action inputs and outputs.

    Describes the type, requirements, and UI representation
    of an action parameter.
    """
    name: str
    type: str  # "string", "number", "boolean", "array", "object"
    required: bool = False
    description: str = ""
    default: Any = None
    ui_component: Optional[str] = None
    options: list[ParamOption] = field(default_factory=list)
    items_type: Optional[str] = None  # For arrays: type of items
    properties: dict[str, "ParamSchema"] = field(default_factory=dict)  # For objects

    def to_dict(self) -> dict:
        """Convert to dictionary for JSON serialization."""
        result = {
            "name": self.name,
            "type": self.type,
            "required": self.required,
            "description": self.description,
        }
        if self.default is not None:
            result["default"] = self.default
        if self.ui_component:
            result["uiComponent"] = self.ui_component
        if self.options:
            result["options"] = [
                {"label": o.label, "value": o.value, "description": o.description}
                for o in self.options
            ]
        if self.items_type:
            result["itemsType"] = self.items_type
        if self.properties:
            result["properties"] = {k: v.to_dict() for k, v in self.properties.items()}
        return result


@dataclass
class ActionExample:
    """
    Example usage for an action.

    Provides sample input/output for documentation and testing.
    """
    name: str
    description: str = ""
    input: dict[str, Any] = field(default_factory=dict)
    config: dict[str, Any] = field(default_factory=dict)
    output: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict:
        """Convert to dictionary for JSON serialization."""
        return {
            "name": self.name,
            "description": self.description,
            "input": self.input,
            "config": self.config,
            "output": self.output,
        }


@dataclass
class ActionMeta:
    """
    Metadata for a plugin action.

    Describes what the action does, its inputs, outputs, and examples.
    This metadata is used by Studio UI for action configuration.
    """
    name: str
    category: str = ""
    description: str = ""
    tags: list[str] = field(default_factory=list)
    inputs: dict[str, ParamSchema] = field(default_factory=dict)
    outputs: dict[str, ParamSchema] = field(default_factory=dict)
    examples: list[ActionExample] = field(default_factory=list)

    def to_dict(self) -> dict:
        """Convert to dictionary for JSON serialization."""
        return {
            "name": self.name,
            "category": self.category,
            "description": self.description,
            "tags": self.tags,
            "inputs": {k: v.to_dict() for k, v in self.inputs.items()},
            "outputs": {k: v.to_dict() for k, v in self.outputs.items()},
            "examples": [e.to_dict() for e in self.examples],
        }


@dataclass
class Manifest:
    """
    Plugin manifest describing identity and capabilities.

    Every plugin must provide a manifest that identifies it
    and declares what it can do.
    """
    name: str
    version: str
    description: str = ""
    author: str = ""
    homepage: str = ""
    license: str = ""
    capabilities: list[str] = field(default_factory=list)

    def to_dict(self) -> dict:
        """Convert to dictionary for JSON serialization."""
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "author": self.author,
            "homepage": self.homepage,
            "license": self.license,
            "capabilities": self.capabilities,
        }


@dataclass
class PatternSuggestion:
    """
    Common pattern suggestion for the analyzer.

    Describes a common usage pattern that the analyzer can suggest
    when similar sequences of actions are detected.
    """
    name: str
    actions: list[str]
    description: str = ""
    benefit: str = ""

    def to_dict(self) -> dict:
        """Convert to dictionary for JSON serialization."""
        return {
            "name": self.name,
            "actions": self.actions,
            "description": self.description,
            "benefit": self.benefit,
        }


@dataclass
class AnalyzerRules:
    """
    Rules for the meta-model analyzer.

    Defines validation rules, action categorizations, and pattern
    suggestions for the plugin's actions.
    """
    plugin_name: str
    plugin_version: str = ""
    action_prefix: str = ""
    provider_type: str = ""
    write_actions: list[str] = field(default_factory=list)
    read_actions: list[str] = field(default_factory=list)
    transaction_action: str = ""
    collection_actions: list[str] = field(default_factory=list)
    collection_field: str = "collection"
    require_schema_prefix: bool = False
    schema_prefix: str = "$schema:"
    failable_actions: list[str] = field(default_factory=list)
    pattern_suggestions: list[PatternSuggestion] = field(default_factory=list)

    def to_dict(self) -> dict:
        """Convert to dictionary for JSON serialization."""
        return {
            "pluginName": self.plugin_name,
            "pluginVersion": self.plugin_version,
            "actionPrefix": self.action_prefix,
            "providerType": self.provider_type,
            "writeActions": self.write_actions,
            "readActions": self.read_actions,
            "transactionAction": self.transaction_action,
            "collectionActions": self.collection_actions,
            "collectionField": self.collection_field,
            "requireSchemaPrefix": self.require_schema_prefix,
            "schemaPrefix": self.schema_prefix,
            "failableActions": self.failable_actions,
            "patternSuggestions": [p.to_dict() for p in self.pattern_suggestions],
        }


# Resource and stats types for plugin status

@dataclass
class PluginResource:
    """Resource tracked by a plugin (provider, connection, etc.)."""
    type: str  # "provider", "connection", "cache", "queue"
    name: str
    status: str  # "connected", "disconnected", "error", "idle"
    info: dict[str, str] = field(default_factory=dict)

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "type": self.type,
            "name": self.name,
            "status": self.status,
            "info": self.info,
        }


@dataclass
class PluginStats:
    """Execution statistics for a plugin."""
    total_executions: int = 0
    successful: int = 0
    failed: int = 0
    avg_latency_ms: float = 0.0
    action_counts: dict[str, int] = field(default_factory=dict)

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "totalExecutions": self.total_executions,
            "successful": self.successful,
            "failed": self.failed,
            "avgLatencyMs": self.avg_latency_ms,
            "actionCounts": self.action_counts,
        }


@dataclass
class PluginStatus:
    """Detailed status of a plugin."""
    manifest: Manifest
    uptime_seconds: int
    resources: list[PluginResource] = field(default_factory=list)
    stats: Optional[PluginStats] = None

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "manifest": self.manifest.to_dict(),
            "uptimeSeconds": self.uptime_seconds,
            "resources": [r.to_dict() for r in self.resources],
            "stats": self.stats.to_dict() if self.stats else None,
        }



"""
Helper functions for App4 plugins.

This module provides utility functions for:
- Config extraction with type safety
- Nested value access/manipulation
- Namespace path handling
- Dictionary operations
"""

from typing import Any, Optional
import copy
import re


def get_config_string(
    config: dict[str, Any], key: str, default: str = ""
) -> str:
    """
    Get a string value from config.

    Args:
        config: Configuration dictionary
        key: Key to look up
        default: Default value if key not found or wrong type

    Returns:
        String value
    """
    value = config.get(key)
    if value is None:
        return default
    if isinstance(value, str):
        return value
    return str(value)


def get_config_bool(
    config: dict[str, Any], key: str, default: bool = False
) -> bool:
    """
    Get a boolean value from config.

    Args:
        config: Configuration dictionary
        key: Key to look up
        default: Default value if key not found or wrong type

    Returns:
        Boolean value
    """
    value = config.get(key)
    if value is None:
        return default
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        return value.lower() in ("true", "1", "yes", "on")
    return bool(value)


def get_config_int(
    config: dict[str, Any], key: str, default: int = 0
) -> int:
    """
    Get an integer value from config.

    Args:
        config: Configuration dictionary
        key: Key to look up
        default: Default value if key not found or wrong type

    Returns:
        Integer value
    """
    value = config.get(key)
    if value is None:
        return default
    if isinstance(value, int) and not isinstance(value, bool):
        return value
    if isinstance(value, float):
        return int(value)
    if isinstance(value, str):
        try:
            return int(value)
        except ValueError:
            return default
    return default


def get_config_float(
    config: dict[str, Any], key: str, default: float = 0.0
) -> float:
    """
    Get a float value from config.

    Args:
        config: Configuration dictionary
        key: Key to look up
        default: Default value if key not found or wrong type

    Returns:
        Float value
    """
    value = config.get(key)
    if value is None:
        return default
    if isinstance(value, (int, float)) and not isinstance(value, bool):
        return float(value)
    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            return default
    return default


def get_config_value(config: dict[str, Any], key: str) -> Any:
    """
    Get a raw value from config.

    Args:
        config: Configuration dictionary
        key: Key to look up

    Returns:
        Raw value or None if not found
    """
    return config.get(key)


def get_config_list(
    config: dict[str, Any], key: str
) -> list[Any]:
    """
    Get a list value from config.

    Args:
        config: Configuration dictionary
        key: Key to look up

    Returns:
        List value or empty list
    """
    value = config.get(key)
    if value is None:
        return []
    if isinstance(value, list):
        return value
    return []


def get_config_dict(
    config: dict[str, Any], key: str
) -> dict[str, Any]:
    """
    Get a dictionary value from config.

    Args:
        config: Configuration dictionary
        key: Key to look up

    Returns:
        Dictionary value or empty dict
    """
    value = config.get(key)
    if value is None:
        return {}
    if isinstance(value, dict):
        return value
    return {}


def require_config(config: dict[str, Any], *keys: str) -> None:
    """
    Ensure required config keys exist.

    Args:
        config: Configuration dictionary
        keys: Required keys

    Raises:
        ValueError: If any required key is missing
    """
    missing = [k for k in keys if k not in config or config[k] is None]
    if missing:
        raise ValueError(f"Missing required config keys: {', '.join(missing)}")


def get_nested_value(data: dict[str, Any], path: str) -> Any:
    """
    Get a nested value using dot notation.

    Args:
        data: Dictionary to search
        path: Dot-separated path (e.g., "user.profile.name")

    Returns:
        Value at path or None if not found

    Examples:
        get_nested_value({"user": {"name": "John"}}, "user.name")  # "John"
        get_nested_value({"items": [1, 2, 3]}, "items.0")  # 1
    """
    if not path:
        return data

    parts = path.split(".")
    current = data

    for part in parts:
        if current is None:
            return None

        # Handle array index
        if isinstance(current, list):
            try:
                idx = int(part)
                if 0 <= idx < len(current):
                    current = current[idx]
                else:
                    return None
            except ValueError:
                return None
        elif isinstance(current, dict):
            current = current.get(part)
        else:
            return None

    return current


def set_nested_value(data: dict[str, Any], path: str, value: Any) -> None:
    """
    Set a nested value using dot notation.

    Args:
        data: Dictionary to modify
        path: Dot-separated path (e.g., "user.profile.name")
        value: Value to set

    Examples:
        data = {}
        set_nested_value(data, "user.name", "John")
        # data = {"user": {"name": "John"}}
    """
    if not path:
        return

    parts = path.split(".")
    current = data

    for i, part in enumerate(parts[:-1]):
        if part not in current:
            # Check if next part is an integer (array index)
            try:
                int(parts[i + 1])
                current[part] = []
            except ValueError:
                current[part] = {}
        current = current[part]

    # Handle final part
    final_part = parts[-1]
    if isinstance(current, list):
        try:
            idx = int(final_part)
            while len(current) <= idx:
                current.append(None)
            current[idx] = value
        except ValueError:
            pass
    elif isinstance(current, dict):
        current[final_part] = value


def is_namespaced_path(path: str) -> bool:
    """
    Check if a path is a namespaced path.

    Namespaced paths start with $.args., $.ctx., or $.return.

    Args:
        path: Path to check

    Returns:
        True if path is namespaced
    """
    return path.startswith("$.args.") or path.startswith("$.ctx.") or path.startswith("$.return.")


def parse_namespace_path(path: str) -> tuple[str, str]:
    """
    Parse a namespaced path into namespace and remaining path.

    Args:
        path: Namespaced path (e.g., "$.args.user.name")

    Returns:
        Tuple of (namespace, remaining_path)
        e.g., ("args", "user.name")

    Raises:
        ValueError: If path is not a valid namespaced path
    """
    if not is_namespaced_path(path):
        raise ValueError(f"Not a namespaced path: {path}")

    # Remove $. prefix
    path = path[2:]

    # Split into namespace and rest
    parts = path.split(".", 1)
    namespace = parts[0]
    remaining = parts[1] if len(parts) > 1 else ""

    return namespace, remaining


def parse_action_name(name: str) -> tuple[str, str]:
    """
    Parse an action name into prefix and action parts.

    Args:
        name: Action name (e.g., "mongo:find")

    Returns:
        Tuple of (prefix, action)
        e.g., ("mongo", "find")

    Raises:
        ValueError: If action name is invalid
    """
    if ":" not in name:
        raise ValueError(f"Invalid action name format: {name} (expected prefix:action)")

    parts = name.split(":", 1)
    return parts[0], parts[1]


def is_valid_action_name(name: str) -> bool:
    """
    Check if an action name is valid.

    Valid format: prefix:action where both parts are non-empty
    and contain only alphanumeric characters, hyphens, and underscores.

    Args:
        name: Action name to validate

    Returns:
        True if valid
    """
    if ":" not in name:
        return False

    parts = name.split(":", 1)
    if len(parts) != 2 or not parts[0] or not parts[1]:
        return False

    pattern = re.compile(r"^[a-zA-Z][a-zA-Z0-9_-]*$")
    return bool(pattern.match(parts[0]) and pattern.match(parts[1]))


def copy_dict(d: dict[str, Any]) -> dict[str, Any]:
    """
    Deep copy a dictionary.

    Args:
        d: Dictionary to copy

    Returns:
        Deep copy of dictionary
    """
    return copy.deepcopy(d)


def merge_dicts(
    base: dict[str, Any], override: dict[str, Any]
) -> dict[str, Any]:
    """
    Deep merge two dictionaries.

    Values in override take precedence. Nested dictionaries are merged
    recursively. Lists are replaced, not merged.

    Args:
        base: Base dictionary
        override: Override dictionary

    Returns:
        Merged dictionary (new instance)
    """
    result = copy.deepcopy(base)

    for key, value in override.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = merge_dicts(result[key], value)
        else:
            result[key] = copy.deepcopy(value)

    return result
